from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
    
import pyqtgraph as pg
from pyqtgraph import opengl 
import numpy as np
from typing import List, Optional
from PySide6 import QtCore, QtGui
from pyqtgraph import functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph import opengl
from PySide6.QtWidgets import QApplication, QMainWindow
from PySide6 import QtWidgets
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
import sys
from PySide6 import QtGui, QtCore
import pyqtgraph.opengl as gl
import numpy as np
from pyqtgraph.opengl import GLTextItem
from OpenGL import GL as opengl
from OpenGL import GLU
from PySide6 import QtGui
from pprint import pprint
import copy
import math
from pyvol_terminal.interfaces.volatility_surface import view_box

np.set_printoptions(suppress=True)


def project_point(azimuth_deg, elevation_deg, distance, point):
    azimuth = np.radians(azimuth_deg)
    elevation = np.radians(elevation_deg)
    
    forward = np.array([
        np.cos(elevation) * np.cos(azimuth),
        np.cos(elevation) * np.sin(azimuth),
        np.sin(elevation)
    ])
    
    up = np.array([0, 0, 1])
    right = np.cross(forward, up)
    if np.linalg.norm(right) < 1e-10:
        right = np.array([1.0, 0.0, 0.0])
    else:
        right /= np.linalg.norm(right)
    
    up = np.cross(right, forward)
    up /= np.linalg.norm(up)
    
    eye = np.array([0, 1, 0]) - distance * forward
    vec = point - eye
    
    x = np.dot(vec, right)
    y = np.dot(vec, up)
    z = np.dot(vec, forward)
    
    if abs(z) < 1e-10:
        z = 1e-10 if z >= 0 else -1e-10
    
    u = x / z
    v = y / z
    return u, v

def get_world_width(ref_azimuth, ref_elevation, ref_distance, ref_segment_pixel_ref_view,
                    new_azimuth, new_elevation, new_distance, new_pixel_width):
    A = np.array([0, 0, 0])
    B = np.array([1, 0, 0])
    
    uA_ref, vA_ref = project_point(ref_azimuth, ref_elevation, ref_distance, A)
    uB_ref, vB_ref = project_point(ref_azimuth, ref_elevation, ref_distance, B)
    D_ref = np.sqrt((uB_ref - uA_ref)**2 + (vB_ref - vA_ref)**2)
    
    uA_new, vA_new = project_point(new_azimuth, new_elevation, new_distance, A)
    uB_new, vB_new = project_point(new_azimuth, new_elevation, new_distance, B)
    D_new = np.sqrt((uB_new - uA_new)**2 + (vB_new - vA_new)**2)
    
    world_width = (new_pixel_width * D_ref) / (ref_segment_pixel_ref_view * D_new)
    return world_width



class WorldSizedText(GLTextItem):
    origin = np.array([0, 1, 0])
    width_in_world_coord_param = 0.1
    world_width_character = 0.014286
    
    
    def __init__(self, anchor_right, *args, **kwargs):
        if anchor_right:
            self.axis_offset=kwargs["pos"][1]
        self.anchor_right=anchor_right
        self._right_anchor_pos=copy.deepcopy(kwargs["pos"])
      #  print(f"\n{kwargs["text"]}")
  #      print(kwargs["font"].pointSize())
        super().__init__(*args, **kwargs)
        


    def setParent(self, parent: View):
        
        super().setParent(parent)
        return 
        print("")
        print("-----------")
        print(f"{self.text}")
        print(f"self.width_character: {self.world_width_character}")
        print(f"len(self.text): {len(self.text)}")
        print(f"pxWidth()_prev: {self.pxWidth()}")
        print(f"worldWidth()_prev: {self.worldWidth()}")

  #      self.maintain_text_world_width()
        print(f"self._right_anchor_pos: {self._right_anchor_pos}")
        print(f"self.font.pointSizeF: {self.font.pointSizeF()}")
        print(f"pxWidth: {self.pxWidth()}")
        
        
        self.update()
        
        if self.anchor_right:
            self.shiftToWidget(parent)


    def shiftToWidget(self, widget):
        return 
        text_width_world = self.worldWidth()
        camera_pos = widget.cameraPositionNumpy()
        current_angle = math.degrees(math.atan2(*(-1*camera_pos[:2])))
        
        #current_angle = widget.opts["azimuth"]
      
        print("")
        
        
        el_x = math.degrees(math.atan2(camera_pos[0], camera_pos[2]))
        el_y = math.degrees(math.atan2(*camera_pos[1:]))
        x, y = self.offset_xy(text_width_world, current_angle, 0)
        pos = self._right_anchor_pos.copy()
        pos[0] = pos[0] + x
        pos[1] = pos[1] + y
        print(f"\nself._right_anchor_pos: {self._right_anchor_pos}")
        print(f"camera: {camera_pos}")
        print(f"azi: {widget.opts["azimuth"]}")
        
        print(f"current_angle: {current_angle}")
        print(f"el_x: {el_x}")
        print(f"el_y: {el_y}")
        print(f"x: {x}")
        print(f"y: {y}")
        print(f"pos: {pos}")

        self.setData(pos = pos)

    @QtCore.Slot()
    def updatePixelSize(self, view_widget: View):
        self.maintain_text_world_width()
        
        print(f"\n{self.text}")
        print(self.font.pointSizeF())
        print(len(self.text))


    @QtCore.Slot()
    def updatePixelSize2(self, view_widget: View):
        target_width = len(self.text) * self.world_width_character
        print("")
        print("-----------")
        print(f"{self.text}")
        metrics = QtGui.QFontMetricsF(self.font)
        rect = metrics.boundingRect(self.text).getRect()
        coords = metrics.boundingRect(self.text).getCoords()
        print(f"rect: {rect}")
        print(f"coords: {coords}")
        print(f"self.font.pointSizeF(): {self.font.pointSizeF()}")
        self.maintain_text_world_width()
        #self.font.setPixelSize(pixelSize)
        #self.setData(font=self.font)
        print(f"update:")
        #self.maintain_text_world_width()
        
        print(f"pxWidth(): {self.pxWidth()}")
        print(f"self.font.pointSizeF(): {self.font.pointSizeF()}")
        print(f"pointWidth(): {self.pointWidth()}")
        print(f"self.worldWidth(): {self.worldWidth()}")
        print(f"self.pxHeight(): {self.pxHeight()}")
        print(f"self.worldWidth(): {self.pxHeight()}")
        
        rect = metrics.boundingRect(self.text).getRect()
        coords = metrics.boundingRect(self.text).getCoords()

        print(f"rect: {rect}")
        print(f"coords: {coords}")
        
        

        if self.anchor_right:
            self.shiftToWidget(view_widget)
    
    def update_size(self, view_widget: View):
        target_width = len(self.text) * self.world_width_character
        
        # Get current metrics and actual text width
        metrics = QtGui.QFontMetricsF(self.font)
        current_px_width = metrics.boundingRect(self.text).width()
        pixel_size_world = view_widget.pixelSize(self._right_anchor_pos)
        current_world_width = current_px_width * pixel_size_world
        
        if current_world_width <= 0:
            return
        
        # Calculate required scaling factor
        scale_factor = target_width / current_world_width
        
        # Scale the font size while maintaining units
        if self.font.pixelSize() > 0:
            # Pixel-based font: scale directly
            new_px = max(1, int(self.font.pixelSize() * scale_factor))
            self.font.setPixelSize(new_px)
        else:
            # Point-based font: scale points
            new_pt = self.font.pointSizeF() * scale_factor
            if new_pt < 0.1:
                new_pt = 0.1
            self.font.setPointSizeF(new_pt)
        
        self.setData(font=self.font)

    def maintain_text_world_width(self):
        current_world_width = self.worldWidth()
        if current_world_width <= 0:
            return
        
        scale_factor = len(self.text) * self.world_width_character / current_world_width
        
        
        
        current_font = self.font
        current_point_size = current_font.pointSizeF()
        new_size = current_point_size * scale_factor

        current_font.setPointSizeF(new_size)
        self.setData(font=current_font)

            
    def maintain_text_world_width2(self):
        current_world_width = self.worldWidth()
        if current_world_width <= 0:
            return
        
        print(f"current_world_width: {current_world_width}")
        scale_factor = len(self.text) * self.world_width_character / current_world_width
        
        dx = current_world_width * (1- scale_factor)
        dy = current_world_width * (1- scale_factor)
        dz = current_world_width * (1- scale_factor)
      #  self.scale(dx,0,0 )
       
        
        print(f"scale_factor: {scale_factor}")
        
        current_font = self.font
        current_point_size = current_font.pointSizeF()
      #  print(f"current_world_width: {current_world_width}")
     #   print(f"current_point_size: {current_point_size}")
        new_size = current_point_size * scale_factor
    #    print(f"new_size: {new_size}")
        if new_size < 0.1: 
            new_size = 0.1
        current_font.setPointSizeF(new_size)
        self.setData(font=current_font)


    def offset_xy(self, txt_width, theta, elevation):
        x = -txt_width * math.cos(math.radians(theta))
        print(f"x_ang: {math.cos(math.radians(theta))}")
        y = txt_width * math.sin(math.radians(theta))
        print(f"y_ang: {math.sin(math.radians(theta))}")
        return x, y
    
    @classmethod
    def euclidean_distance(cls, point1, point2):
        return math.sqrt((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2 + (point2[2] - point1[2]) ** 2)
    """
    def pxHeight(self):
        metrics = QtGui.QFontMetricsF(self.font)
        return metrics.boundingRect(self.text).height()
    
    def pxWidth(self):
        metrics = QtGui.QFontMetricsF(self.font)
        return metrics.boundingRect(self.text).width()
    
    def pointWidth(self):
        return self.pxWidth() * 72 / self.view().screen().logicalDotsPerInch()

    def worldWidth(self):
        return self.pxWidth() * self.view().pixelSize(self._right_anchor_pos)

    def worldWidthPoint(self):
        return self.pointWidth() * self.view().pointSize(self._right_anchor_pos)

    """

    def pxHeight(self):
        metrics = QtGui.QFontMetricsF(self.font)
        return metrics.boundingRect(self.text).height()
    
    def pxWidth(self):
        metrics = QtGui.QFontMetricsF(self.font)
        return metrics.boundingRect(self.text).width()
    
    def pointWidth(self):
        return self.pxWidth() * 72 / self.view().screen().logicalDotsPerInch()

    def worldWidth(self):
        return self.pxWidth() * self.view().pixelSize(np.array(self.pos))

    def worldWidthPoint(self):
        return self.pointWidth() * self.view().pointSize(np.array(self.pos))



class AxisItem3D:
    def __init__(self, axis_direction, label, colour, axis_pos_vector, axis_origin, tick_direction, tick_size_major, axRange=None, offset_label=0.05, offset_tick_num=None, font=None, parent=None):
        self.axis_direction=axis_direction
        self.label=label
        self.colour=colour
        self._nTicks = axis_pos_vector.size
        self.offset_label=offset_label
        self.font=font if not font is None else QtGui.QFont('Neue Haas Grotesk',  pointSize=12)
        #self.font.setPixelSize(12)
 #      self.font.setPointSizeF(12)
        self.parent=parent
        self.axis_vector = np.zeros(3)
        self.axis_vector[axis_direction]=1
        self.axis_origin=axis_origin
        tick_axis_gap = tick_size_major / 3
        self._tickBufferLinePos = None
        self.tickBufferLine = self._initTickBufferLine(axis_pos_vector, self.axis_vector, self.axis_origin, tick_direction, tick_axis_gap)
        self._tickItemMajor = self._initTickMajorItems(axis_pos_vector, self.axis_vector, self._tickBufferLinePos[0], tick_direction, tick_size_major)
        
        dMajor = axis_pos_vector[1] - axis_pos_vector[0]
        self._tickItemMinor = self._initTickMinorItems(self._tickItemPosMajor, dMajor)
        self.tick_size_major=tick_size_major
        if not axRange is None:
            self.tick_strings=[f"{int(round((i / (1+ 0*self.axis_direction)), 1))}" for i in np.linspace(np.amin(axRange), np.amax(axRange), self._nTicks)]
        #    if axis_direction == 0:
        #        for idx, s in enumerate(self.tick_strings):
          #          s = s + "...."
          #          self.tick_strings[idx] = s
                
        else:
            self.tick_strings=[f"{i}" for i in np.arange(self._nTicks)] #* self._nTicks
        self.tick_values=np.full(self._nTicks, np.nan)
        
        self.tickNumPositions = None
        self.tickNums = self._initTickNums()
        
        self.label = self._initLabelItem(label, axis_pos_vector, self.axis_vector, self.axis_origin, tick_direction, self.offset_label)
        self._ticks_visible=False

    def reset_numItem_pxWidth222(self):
        pointPxDratio=[]
        for textNum in self.tickNums:
            pointPxDratio.append(textNum._pointPxDistanceRatio_default)
        maxPointPxDratio = np.amax(pointPxDratio)
        for textNum in self.tickNums:
            textNum._pointPxDistanceRatio_default=maxPointPxDratio
            
    
    def _initTickBufferLine(self, tick_values, axis_vector, axis_origin, tick_direction, tick_axis_gap):
        tick_values = np.asarray(tick_values)
        axis_vector = np.asarray(axis_vector)
        axis_origin = np.asarray(axis_origin)
        tick_direction = np.asarray(tick_direction)

        t_min = tick_values.min()
        t_max = tick_values.max()

        start = axis_origin + t_min * axis_vector + tick_direction * tick_axis_gap
        end = axis_origin + t_max * axis_vector + tick_direction * tick_axis_gap
        self._tickBufferLinePos=np.vstack((start, end))
        return gl.GLLinePlotItem(pos=self._tickBufferLinePos, color="white", width=2.5)

    def _initTickMajorItems(self, tick_values, axis_vector, shifted_axis_origin, tick_direction, tick_size):
        tick_values = np.asarray(tick_values)
        axis_vector = np.asarray(axis_vector)
        tick_direction = np.asarray(tick_direction)
        shifted_axis_origin = np.asarray(shifted_axis_origin)
        starts = shifted_axis_origin + np.outer(tick_values, axis_vector)
        ends = starts + tick_size * tick_direction
        self._tickItemPosMajor = [np.vstack((start, end)) for start, end in zip(starts, ends)]
        return [gl.GLLinePlotItem(pos=coord, color="white", width=2.5) for coord in self._tickItemPosMajor]

        
    def _initTickMinorItems(self, coordinates_major, dMajor):
        
        self._tickItemPosMinor = copy.deepcopy(coordinates_major[1:])
        tick_items =[]
        
        for coordinate in self._tickItemPosMinor:
            dSizeMajor = coordinate[1] - coordinate[0]
            coordinate[1] = coordinate[0] + dSizeMajor / 2
            coordinate[:, self.axis_direction] -= dMajor / 2

            tick = gl.GLLinePlotItem(pos=coordinate, color="white", width=1.5)
            tick_items.append(tick)
        return tick_items


    def setParent(self, parent):
        self.parent=parent
        for idx, tickNum in enumerate(self.tickNums):
            if hasattr(tickNum, "updatePixelSize"):
                self.parent.viewChangedSignal.connect(tickNum.updatePixelSize)
            self.parent.addItem(tickNum, viewBoxAdd=False)
            tickNum.setParent(self.parent)       
        
        for tick in self._tickItemMajor + self._tickItemMinor:
            tick.setParent(self.parent)
            parent.addItem(tick, viewBoxAdd=False)
  #      self.parent.addItem(self.label, viewBoxAdd=False)
   #     self.label.setParent(self.parent)
   #     if hasattr(self.label, "updatePixelSize"):
    #        self.parent.viewChangedSignal.connect(self.label.updatePixelSize)
        self._ticks_visible=True
        self.parent.addItem(self.tickBufferLine)

        
    #    self.reset_numItem_pxWidth()
        
    
    def _initTickNums(self) -> List[WorldSizedText]:
        gl_objects = []
        if self.axis_direction == 0:
            anchor_right=True
        else:
            anchor_right=False
        idx=0
        print(f"len(self.tick_strings)-1: {len(self.tick_strings)-1}")
        for pos_vec, tick_str in zip(self._tickItemPosMajor, self.tick_strings):
            print(idx)
            if not idx in [0, len(self.tick_strings)-1]:
                idx+=1
                continue 
            idx+=1
            pos = pos_vec[1,:]

            gl_object = WorldSizedText(anchor_right, pos=pos, text=tick_str, font=self.font, color=self.colour)
        #   gl_object = GLTextItem(pos=pos, text=tick_str, font=self.font, color=self.colour)
            gl_objects.append(gl_object)
            
        return gl_objects

    def create_tick_strings(self):
        return [str(val) for val in self.tick_values]
    
    def create_tick_values(self, min_val, max_val, n_ticks):
        return np.linspace(min_val, max_val, n_ticks)

    def setLabel(self, label):
        self._label=label
        self.label.setData(text=label)
    
    def getLabel(self):
        return self._label
      
    @QtCore.Slot()
    def update_values(self, plot_view):
        self.min, self.max = plot_view.min, plot_view.max
        self.tick_values = self.create_tick_values(self.min, self.max, self._nTicks)
        self.tick_strings = self.create_tick_strings()
        for tick_str, pos_vec, tick_object in zip(self.tick_strings, self._tickItemPosMajor, self.tickNums):
            pos = pos_vec[1,:]
            if not self._ticks_visible:
                tick_object.show()
            try:
                tick_object.setData(text=tick_str, pos=pos)      
            except:
                pass
        if not self._ticks_visible:
            self._ticks_visible=True

    def _initLabelItem(self, label, tick_values, axis_vector, axis_origin, tick_direction, label_offset):
        tick_values = np.asarray(tick_values)
        axis_vector = np.asarray(axis_vector)
        axis_origin = np.asarray(axis_origin)
        tick_direction = np.asarray(tick_direction)

        t_min = tick_values.min()
        t_max = tick_values.max()
        t_mid = 0.5 * (t_min + t_max)

        midpoint = axis_origin + t_mid * axis_vector
        offset = label_offset * tick_direction
        if self.axis_direction == 0:
            anchor_right=True
        else:
            anchor_right=False
        return GLTextItem(pos=midpoint + offset, text=label, font=self.font, color=self.colour)#WorldSizedText(anchor_right, pos=midpoint + offset, text=label, font=self.font, color=self.colour)


class CustomGrid(gl.GLGridItem):
    def __init__(self, n_lines, *args, **kwargs):
        self.n_lines=int(2*n_lines -1)
        super().__init__(*args, **kwargs)

    def updateLines(self):
        if self.lineplot is None:
            return
        x, y, _ = self.size()
        xvals = np.linspace(-x/2., x/2., self.n_lines)
        yvals = np.linspace(-y/2., y/2., self.n_lines)
        
        set1 = np.zeros((len(xvals), 6), dtype=np.float32)
        set1[:, 0] = xvals
        set1[:, 1] = yvals[0]
        set1[:, 3] = xvals
        set1[:, 4] = yvals[-1]

        set2 = np.zeros((len(yvals), 6), dtype=np.float32)
        set2[:, 0] = xvals[0]
        set2[:, 1] = yvals
        set2[:, 3] = xvals[-1]
        set2[:, 4] = yvals

        pos = np.vstack((set1, set2)).reshape((-1, 3))
        self.lineplot.setData(pos=pos, color=self.color())
        self.update()


class CustomAxisItem(pg.AxisItem):
    def __init__(self, axis_direction=None, tick_engine=None, *args, **kwargs):
        self.axis_direction=axis_direction
        self.title=None
        self.tick_engine=tick_engine
        self.global_count=0
        super().__init__(*args, **kwargs)
        self.enableAutoSIPrefix(False)

    def tickStrings(self, values, scale, spacing):
        return self.tick_engine.function(values)      
    
class AxisManager(QtCore.QObject):
    def __init__(self,
                 view_widget,
                 n_major_ticks=[5] * 3
                 ):
        super().__init__()
        self.view_widget=view_widget
        self.str_int_map = {"x" : 0,
                            "y" : 1,
                            "z" : 2}
        self.int_str_map = {value : key for key, value in self.str_int_map.items()}
        self.n_ticks_dict={0 : n_major_ticks[0],
                           1 : n_major_ticks[1],
                           2 : n_major_ticks[2]}
        self.axis_2D_items: dict[str, List[CustomAxisItem]] = {0 : [],
                                                               1 : [],
                                                               2 : []}              
        self.axis_3D_items: dict[str, None] = {0 : None,
                                                     1 : None,
                                                     2 : None}
        self.colours = {0 : "white",
                        1 : "white",
                        2 : "white",
                        }
        self.axis_x = None  
        self.axis_y = None
        self.axis_z = None
        self.labels=["Strike", "Expiry", "Implied Volatility"]        
        self.offset_ticks = {0 : np.array([-0.2, 0, 0]),
                             1 : np.array([0, 0.2, 0]),
                             2 : np.array([0.2, 0.2, 0]),
                             }
        
        
        self._base_label_positions = {0 : np.array([0.5, 0, 0]),
                                      1 : np.array([1, 0.5, 0]),
                                      2 : np.array([1, 1, 0.5])}
        self.base_ticknum_positions = {}
        self.offset_tick_num = {jdx : np.zeros(3) for jdx in range(3)}
        self.tick_size_major=1/15
        axis_pos = np.linspace(0, 1, self.n_ticks_dict[0])
        self.tick_size_structure={0 : np.array([0, -1, 0]),
                                  1 : np.array([1, 0, 0]),
                                  2 : np.array([1, 1, 0]),
                                  }        
        self.axis_pos_vector = {jdx : np.linspace(0, 1, self.n_ticks_dict[jdx]) for jdx in range(3)}

        self.axis_origin_container = {0 : (0, 0, 0),
                                      1 : (1, 0, 0),
                                      2 : (1, 1, 0)
                                      }
        self.tick_direction_container = {0 : (0, -1, 0),
                                         1 : (1, 0, 0),
                                         2 : (1, 0, 0)}
        
        self.base_axis_coordinates={0 : [np.array([idx, 0, 0]) for idx in axis_pos],
                                    1 : [np.array([1, idx, 0]) for idx in axis_pos],
                                    2 : [np.array([1, 1, idx]) for idx in axis_pos],
                                    }        
        
        self.offset_labels = {0 : 0.2,
                              1 : 0.2,
                              2 : 0.2,
                              }
        self._init_axis_3D_items(view_widget)
        self.initialised_default=True

    def _init_axis_3D_items(self, view_widget):
        Range = np.array([view_widget.min, view_widget.max])
        for axis_direction, label in zip([0, 1, 2], self.labels):
       #     if axis_direction != 0: 
               # continue
            print(f"axis: {axis_direction}")
            view_range = Range * (1 + axis_direction*0.1)
            if axis_direction > 1:
                break
            axis_str = self.int_str_map[axis_direction]
            axis_item = AxisItem3D(axis_direction, 
                                    label,
                                    self.colours[axis_direction],
                                    self.axis_pos_vector[axis_direction],
                                    self.axis_origin_container[axis_direction],
                                    self.tick_direction_container[axis_direction],
                                    self.tick_size_major,
                                    view_range,
                                    self.offset_labels[axis_direction],
                                    self.offset_tick_num[axis_direction])
            
            if not view_widget is None:
                view_widget.addItem(axis_item)
                axis_item.setParent(view_widget)
                view_widget.axisChangedSignal.connect(axis_item.update_values)
                
            self.axis_3D_items[axis_direction] = axis_item
        view_widget.update()

    def get_label(self,
                  axis_direction: str
                  ) -> str:
        return self.axis_3D_items[axis_direction].label.getLabel()

    def switch_axis(self, axis_direction, axis_label):
        if isinstance(axis_direction, str):
            axis_direction=self.str_int_map[axis_direction]
        self.axis_3D_items[axis_direction].setLabel(axis_label)  
        for axis_item in self.axis_2D_items[axis_direction]:
            axis_item.setLabel(axis_label)
            
    def add_2D_axis_item(self, axis_item, axis_direction):
        self.axis_2D_items[axis_direction].append(axis_item)
        
    def addWidget(self, widget):
        self.view_widget=widget
        if not self.initialised_default:
            self._init_axis_3D_items()
        for axis_item in self.axis_3D_items.values():
            widget.axisChangedSignal.connect(axis_item.update_values)
    

class GridManager:
    def __init__(self,
                 widget: opengl.GLViewWidget,
                 n_major_ticks: int
                 ) -> GridManager:
        self.widget=widget
        self.size=[1] * 3
        self.n_major_ticks=n_major_ticks
        self.n_minor_ticks=[2 * t  for t in n_major_ticks]
        print(f"self.n_minor_ticks: {self.n_minor_ticks}")
        self.spacing= [size / minor_tick for size, minor_tick in zip(self.size, self.n_minor_ticks)]
        self._major_minor_scale = 0.5
        
        self.grid_xy = self._create_grid(self.size, self.spacing, translation=(0.5, 0.5, 0))
        
        offset_yz = -0.3 
        offset_xz = 0.3
        
        self.grid_yz = self._create_grid(self.size, self.spacing, rotation=(90, 0, 1, 0), translation=(0 + offset_yz, 0.5, 0.5))
        self.grid_xz = self._create_grid(self.size, self.spacing, rotation=(90, 1, 0, 0), translation=(0.5, 1 + offset_xz, 0.5))
        if not widget is None:
            self._addGrids(widget)
    
    def addWidget(self,
                  widget: opengl.GLViewWidget
                  ) -> None:
        self.widget=widget
        self._addGrids(self.widget)
    
    def _addGrids(self,
                  widget: opengl.GLViewWidget
                  ) -> None:
        widget.addItem(self.grid_xy)
        widget.addItem(self.grid_yz)
        widget.addItem(self.grid_xz)

        self.widget=widget
        
    def _create_grid(self, size, spacing, rotation=None, translation=None):      
        print(f" self.n_major_ticks[0]: {self.n_major_ticks[0]}")  
        grid = CustomGrid(n_lines=self.n_major_ticks[0])
        grid.setSize(*size)
        grid.setSpacing(*spacing)
        if rotation:
            grid.rotate(*rotation)        
        if translation:
            grid.translate(*translation)
        return grid

class View(gl.GLViewWidget):
    rc_pos_unnorm_signal = QtCore.Signal(float, float, float)
    mouse_position_signal = QtCore.Signal(float, float, float)
    viewChangedSignal = QtCore.Signal(object)
    axisChangedSignal = QtCore.Signal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vb=view_box.ViewBox()

        self.setMouseTracking(True)
        self.setAttribute(QtCore.Qt.WA_Hover)
        self.setAttribute(QtCore.Qt.WA_DontCreateNativeAncestors) 
        self.show_spot_text=True
        near_clip, far_clip = self.compute_optimal_clipping()
        self._mouseZoomFactor=0.999
        distance = 4
        azimuth = -0#-45 #+ math.degrees(math.asin(1/3))#-90
        elevation = 15#math.degrees(math.acos(math.sqrt(2)/4))#0#30
        self._bounds_cartesian = [[0, 1], [0, 1], [0, 1]]
        self.cross_hairs_enabled=True

        self.opts['near'] = near_clip
        self.opts['far'] = far_clip
        self.opts["distance"] = distance
        self.opts['azimuth'] = azimuth
        self._azimuth_offset = 90
        self._elevation_offset = 90
        
        self.opts["elevation"] = elevation
        self.opts["center"] = QtGui.QVector3D(0, 1, 0)

        self._spherical_bounds = {"azimuth" : [-90, 0],
                                  "elevation" : [0, 90]}
        self.prev_euler_coords = {"distance" : distance,
                                  "azimuth" : azimuth,
                                  "elevation" : elevation}
        self._cartesian_view_coords = np.array([0., 0., 0.])
        self._cartesian_view_coords = self.cameraPostionVector()
        self.interacting=False
        self.min = 100000.
        self.max = 120000.#95000#125000.
        self.setMouseTracking(True)
        self.update()    

    def pointSize(self, pos):
        PIXELS_PER_POINT = self.screen().logicalDotsPerInch() / 72.0
        
        # Get the pixel size at the given position
        pixel_size = self.pixelSize(pos)
        
        # Calculate the point size
        point_size = pixel_size * PIXELS_PER_POINT
        
        return point_size

    def compute_optimal_clipping(self):
        largest_point_distance = np.sqrt(3)
        near_clip = max(0.0001, self.opts['distance'] - 1.01 * largest_point_distance) 
        far_clip = self.opts['distance'] 
        return near_clip, far_clip

    def setAxisItems(self, axisItems) -> None:
        for axis_item in axisItems.values():
            axis_item.setParent(self)
            self.addItem(axis_item)
            
    def mousePressEvent(self, event):
        match event.buttons().value :    
            case 1: 
                self.interacting=True
                
            case 2:
                self.mouse_pos = event.pos()  
                self.opts["azimuth"] = -14.477512185929925
                
        #        self.viewChangedSignal.emit(self)
                #self.process_right_click(self.mouse_pos)
        super().mousePressEvent(event)

    def process_right_click(self, mouse_pos):
        return 
        world_x, world_y, world_z = self.get_mouse_pos(mouse_pos)
        if 0 <= world_x <= 1 and 0 <= world_y <= 1 and 0 <= world_z <= 1:        
            self.prev_x, self.prev_y, self.prev_z = self.vb.plot_views["x"].unnormalise(world_x),\
                                                    self.vb.plot_views["y"].unnormalise(world_y),\
                                                    self.vb.plot_views["z"].unnormalise(world_z)
            self.rc_pos_unnorm_signal.emit(self.prev_x, self.prev_y, self.prev_z)

    def cameraPositionNumpy(self):
        p = self.cameraPosition()
        return np.array((p.x(), p.y(), p.z()))

    def mouseMoveEvent(self, ev):
        self.mouse_pos = ev.pos()
        if ev.buttons() == QtCore.Qt.MouseButton.LeftButton:
            self.interacting=True

        lpos = ev.position() if hasattr(ev, 'position') else ev.localPos()
        if not hasattr(self, 'mousePos'):
            self.mousePos = lpos
        diff = lpos - self.mousePos
        self.mousePos = lpos
        
        if ev.buttons() == QtCore.Qt.MouseButton.LeftButton:
            self.orbit(-diff.x(), diff.y())
        elif ev.buttons() == QtCore.Qt.MouseButton.MiddleButton:
            if (ev.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier):
                self.pan(diff.x(), 0, diff.y(), relative='view-upright')
            else:
                self.pan(diff.x(), diff.y(), 0, relative='view-upright')
    def addItem(self, item, *args, **kwargs):
        vbargs = {}
        added=False
        if 'ignoreBounds' in kwargs:
            vbargs['ignoreBounds'] = kwargs['ignoreBounds']
        
        if "viewBoxAdd" in kwargs:
            if kwargs["viewBoxAdd"]:
                self.vb.addItem(item, *args, **vbargs)
                self._internal_id_item_map[item.get_internal_id()] = item
            else:
                added=True
                del kwargs["viewBoxAdd"]
                super().addItem(item, *args, **kwargs)
        if not isinstance(item, AxisItem3D) and "viewBoxAdd" in kwargs:
            del kwargs["viewBoxAdd"]
            if not added:
                added=True
                super().addItem(item, *args, **kwargs)
        else:
            if not added and not isinstance(item, AxisItem3D):
                added=True
                super().addItem(item, *args, **kwargs)
        return item


    def orbit(self, azim, elev):    
        """
        if self.opts['azimuth'] + azim < -90 or 0 < self.opts['azimuth'] + azim:
            return 
        if fn.clip_scalar(self.opts['elevation'] + elev, -90., 90.)  < 0:
            return
        """
        super().orbit(azim, elev)
        print(f"\n\nnew_orb\n")
        self.viewChangedSignal.emit(self)
        return
        elevation = fn.clip_scalar(self.opts['elevation'] + elev, -90., 90.) 
        new_opts = {"elevation" : elevation,
                    "azimuth" : self.opts['azimuth'] + azim,
                    "distance" : self.opts["distance"]} 
        
        if not self._validAngle(self.opts['azimuth'] + azim, elevation):
            return
        
        else:
            if self._outsideBox(new_opts):
                cartesian_coords = self.cameraPostionVector(new_opts)
                self.opts.update(new_opts)
                self.update()
                self.viewChangedSignal.emit(self)
            else:
                dE = elevation - self.opts["elevation"]
                cartesian_coords = self.cameraPostionVector(self.opts)
                prev_axis_outside_cube = [self._bounds_cartesian[idx][1] < pos for idx, pos in enumerate(cartesian_coords)]
                azim/=2 
                azimuth = self.opts['azimuth'] + azim
                elevation = self.opts["elevation"] + dE/2
                
                if sum(prev_axis_outside_cube) > 0:
                    for axis, flag in zip("xyz", prev_axis_outside_cube):
                        if flag:
                            if axis == "x":
                                r_new = 1.00001 * self._bounds_cartesian[0][1] / (math.sin(math.radians(self._elevation_offset - elevation))
                                                                       *math.cos(math.radians(azimuth + self._azimuth_offset)))
                                
                            elif axis == "y":
                                r_new = 1.00001 * self._bounds_cartesian[1][1] / (math.sin(math.radians(self._elevation_offset - elevation))
                                                                    *math.sin(math.radians(azimuth + self._azimuth_offset)))
                                
                            else:
                                r_new = 1.00001 * self._bounds_cartesian[2][1] / math.cos(math.radians(self._elevation_offset - elevation))
                    
                    new_opts = {"elevation" : elevation,
                                "azimuth" : azimuth,
                                "distance" : r_new} 
                    self.opts.update(new_opts)

                    self.update()
                    self.viewChangedSignal.emit(self)


    def pan2(self, dx, dy, dz, relative='global'):
        if relative == 'view':
            elev = math.radians(self.opts['elevation'])
            azim = math.radians(self.opts['azimuth'])
            fov = math.radians(self.opts['fov'])
            dist = (self.opts['center'] - self.cameraPosition()).length()
            fov_factor = math.tan(fov / 2) * 2
            scale_factor = dist * fov_factor / self.width()
            z = scale_factor * math.cos(elev) * dy
            x = scale_factor * (math.sin(azim) * dx - math.sin(elev) * math.cos(azim) * dy)
            y = scale_factor * (math.cos(azim) * dx + math.sin(elev) * math.sin(azim) * dy)

            if self._validViewCartesian((x, -y, z)):
                self.opts['center'] += QtGui.QVector3D(x, -y, z)
                self.opts.update({"elevation" : elev,
                                  "azimuth" : azim,
                                  "distance" : dist})
                self.update()
        else:
            raise ValueError("relative argument must be global, view, or view-upright")

    def zoom(self, rMultiplier):
        distance = self.opts['distance'] * rMultiplier
        opts = {"distance" : distance,
                "elevation" : self.opts["elevation"],
                "azimuth" : self.opts["azimuth"]}
        
        if self._outsideBox(opts):
            self.opts["distance"] = distance
            self.update()
            self.viewChangedSignal.emit(self)
            return True
        else:
            return False
            
    
    def wheelEvent(self, ev):
        delta = ev.angleDelta().x()
        if delta == 0:
            delta = ev.angleDelta().y()
        for i in [1, 1/4]:
            zoomChanged=self.zoom(self._mouseZoomFactor**(i*delta))
            if zoomChanged:
                return 
        
    def cameraPostionVector(self,
                            opts=None) -> np.ndarray[float]:
        if opts is None:
            opts = self.opts

        azimuth = opts["azimuth"] + self._azimuth_offset
        elevation = self._elevation_offset - opts["elevation"]
        self._cartesian_view_coords[0] = opts["distance"] * math.sin(math.radians(elevation)) * math.cos(math.radians(azimuth))
        self._cartesian_view_coords[1] = opts["distance"] * math.sin(math.radians(elevation)) * math.sin(math.radians(azimuth))
        self._cartesian_view_coords[2] = opts["distance"] * math.cos(math.radians(elevation))
        return self._cartesian_view_coords

    def _checkCartesianBounds(self, coords):
        return [self._bounds_cartesian[i][0] <= coords[i] < self._bounds_cartesian[i][1] for i in range(3)]
    
    def _outsideBox(self, opts):
        if any([self._bounds_cartesian[idx][1] <= pos for idx, pos in enumerate(self.cameraPostionVector(opts))]):
            return True
        else:
            return False
        
    def _validAngle(self, azimuth, elevation):
        if (self._spherical_bounds["azimuth"][0] <= azimuth <= self._spherical_bounds["azimuth"][1]
            and self._spherical_bounds["elevation"][0] <= elevation <= self._spherical_bounds["elevation"][1]):
            return True
        else:
            return False
    
    def _validView(self, opts):
        cartesian = self.cameraPostionVector(opts)
        
        if any([pos < self._bounds_cartesian[idx][0] or pos > self._bounds_cartesian[idx][1] for idx, pos in enumerate(cartesian)]):
            return False
        else:
            return True
        
    
    @classmethod
    def distanceSphericalCoordinates(cls, r, r_p):
        t1 = r["distance"] **2 + r_p["distance"] ** 2
        
        t21 = (math.sin(math.radians(90 - r["elevation"]))
                *math.sin(math.radians(90 - r_p["elevation"]))
                *math.cos(math.radians(r["azimuth"] - r_p["azimuth"])))
        t22 = (math.cos(math.radians(90 - r["elevation"]))
              *math.cos(math.radians(90 - r_p["elevation"])))
        
        return math.sqrt(t1 - 2 * r["distance"] * r_p["distance"] * (t21 + t22))




class MainWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('Volatility Surface')
        self.widget_central = QtWidgets.QWidget()
        self.setCentralWidget(self.widget_central)
        self.layout_main = QtWidgets.QVBoxLayout()
        self.widget_central.setLayout(self.layout_main)
        self.showMaximized()

        self.view_widget = View()
        self.vw_layout = QtWidgets.QVBoxLayout()
        self.layout_main.addLayout(self.vw_layout)
        
        n_major_ticks = [5]* 3
        
        self.layout_main.addWidget(self.view_widget)
        
        self.grid_manager = GridManager(self.view_widget,
                                        n_major_ticks
                                        )
        
        # Show window then initialize axes with delay
        self.showMaximized()
        QtCore.QTimer.singleShot(100, self.init_axes_after_show)
        
    def init_axes_after_show(self):
        """Initialize axis system after window is shown and sized"""
        n_major_ticks = [5] * 3
        self.axis_manager = AxisManager(self.view_widget, n_major_ticks)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    mainWin = MainWindow()
    sys.exit(app.exec())
    



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
from dataclasses import dataclass, field
import numpy as np

from ..engines.option_engines.base import ABCEuroEngine
from abc import ABC, abstractmethod


__GREEKS__ = {"delta" : np.nan,
              "gamma" : np.nan,
              "vega" : np.nan,
              "theta" : np.nan,
              "rho" : np.nan,
              }

__MONEYNESS__ = {"moneyness" : np.nan,
                 "log_moneyness" : np.nan,
                 "standardised_moneyness" : np.nan,
                 "forward_moneyness" : np.nan,
                 "discounted_moneyness" : np.nan,
                 }


@dataclass(slots=True)
class ABCMetric(ABC):
    targets: List[str] = field(default_factory=list)
    
    def get_targets(self): return self.targets
    
    def add_target(self, target):
        self.targets.append(target)
    
    def remove_target(self, target):
        self.targets.remove(target)
    
    @abstractmethod
    def update(self, *args, **kwargs): ...
    
    


@dataclass(slots=True, frozen=True, kw_only=True)
class IVOLConfig:
    px_type_option: str
    px_type_underlying: str
    



        
@dataclass(slots=True)
class Greeks(ABCMetric):
    delta: float=np.nan
    gamma: float=np.nan
    vega: float=np.nan
    theta: float=np.nan
    rho: float=np.nan
    
    def update(self, ivol):
        self.ivol=ivol


@dataclass(slots=True, kw_only=True)
class Moneyness(ABCMetric):
    moneyness: float=np.nan
    log_moneyness: float=np.nan
    standardised_moneyness: float=np.nan
    forward_moneyness: float=np.nan
    discounted_moneyness: float=np.nan
        
    
    
@dataclass(slots=True, kw_only=True)
class IVOLMetrics:
    
    """
    convenience class to help manage metrics that need to be calculated and prevent double calculation etc 

    """
    
    config: IVOLConfig
    last_ts: float=np.nan
    ivol: float=np.nan
    greeks: Greeks = field(default_factory=lambda : __GREEKS__.copy())
    moneyness: Moneyness = field(default_factory=lambda : __MONEYNESS__.copy())
    
    _cachedParams: Dict[str, float] = field(default_factory=dict)
    _cacheTsTol=4.

    def get_req_metrics(self):
        return self.greeks.targets + self.moneyness.targets 
    
    def set_metrics(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self.greeks, k, v)
        
    def update(self, ivol, cache: Dict[str, float]=None):
        self.ivol=ivol
        if not cache is None:
            self._cachedParams = cache.copy()

    def is_cached(self,
                  update_option: Tuple[float, float]=None,
                  update_under: Tuple[float, float]=None,
                  ) -> bool:
        
        if not update_option is None:
            return self._cacheTsTol > update_option[0] - self._cachedParams["option"][0] \
                and update_option[1] == self._cachedParams["option"][1]
        if not update_under is None:
            return self._cacheTsTol > update_under[0] - self._cachedParams["underlying"][0] \
                and update_under[1] == self._cachedParams["underlying"][1]

    
@dataclass(slots=True, kw_only=True)
class GreeksN(Greeks):
    Ndelta: float=np.nan
    




@dataclass(slots=True, kw_only=True)
class OptionState:
    constants: Dict[str, str|float]
    OTM: bool=None
    
    def check_set_get(self, attr, px_underlying):
        if attr == "OTM":
            if self.constants["flag"] == "c":
                self.OTM = px_underlying < self.constants["strike"]
            else:
                self.OTM = px_underlying > self.constants["strike"]            
            return self.OTM
    

@dataclass(slots=True, kw_only=True)
class BarrierOptionState(OptionState):
    knocked_in: bool=None



from pyqtgraph.Qt import QtCore
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph.Qt import QtCore, QtGui
from PySide6 import QtWidgets
from pyqtgraph.Qt import QT_LIB
import weakref
from pyqtgraph import opengl
import numpy as np
import warnings
from pyqtgraph.Qt import isQObjectAlive
from pyvol_terminal.gl_3D_graphing import meta
import abc

        

class WeakList(object):

    def __init__(self):
        self._items = []

    def append(self, obj):
        #Add backwards to iterate backwards (to make iterating more efficient on removal).
        self._items.insert(0, weakref.ref(obj))

    def __iter__(self):
        i = len(self._items)-1
        while i >= 0:
            ref = self._items[i]
            d = ref()
            if d is None:
                del self._items[i]
            else:
                yield d
            i -= 1

class AbstractGLGraphicsItem(GLGraphicsItem, metaclass=meta.QABCMeta):
    
    def __init__(self, parentItem=None, **kwargs):
        self._blockUpdate: bool
        self.blockUpdate(False)
        self._viewBox=None
        self._viewWidget=None
        super().__init__()

        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        self._cachedView = None
        self.setParentItem(parentItem)

    def blockUpdate(self, flag):
        self._blockUpdate=flag
            
        
    @abc.abstractmethod
    def _internal_update(self):...

    def update(self):
        if not self._blockUpdate:
            self._internal_update()
        
    
    def setParentItem(self, item):
        if self._GLGraphicsItem__parent is not None:
            self._GLGraphicsItem__parent._GLGraphicsItem__children.remove(self)
        if item is not None:
            item._GLGraphicsItem__children.append(self)
            item.itemChange(item.GraphicsItemChange.ItemChildAddedChange, self)

        if self._GLGraphicsItem__view is not None:
            self._GLGraphicsItem__view.removeItem(self)

        self._GLGraphicsItem__parent = item
        self._GLGraphicsItem__view = None
        

    def childItems(self):
        return list(self._GLGraphicsItem__children)


class GLItemGroup(AbstractGLGraphicsItem, QtWidgets.QGraphicsObject):
    def __init__(self, parentItem=None):
        self.__inform_view_on_changes = True
        super().__init__(parentItem=parentItem)
        
        self.viewRange = [[0, 1] for _ in range(3)]
        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        self.itemsChangedListeners = WeakList()
        
        
        self.setFlag(self.GraphicsItemFlag.ItemHasNoContents)
        self.setFlag(self.GraphicsItemFlag.ItemSendsGeometryChanges)
        self.setParentItem(parentItem)

    def itemChangeInner(self, change, value):
        ret = super().itemChange(change, value)
        if change in [self.GraphicsItemChange.ItemParentHasChanged, self.GraphicsItemChange.ItemSceneHasChanged]:
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                self.changeParent()
        try:
            inform_view_on_change = self.__inform_view_on_changes
        except AttributeError:
            # It's possible that the attribute was already collected when the itemChange happened
            # (if it was triggered during the gc of the object).
            pass
        else:
            if inform_view_on_change and change in [self.GraphicsItemChange.ItemPositionHasChanged, self.GraphicsItemChange.ItemTransformHasChanged]:
                self.informViewBoundsChanged()
        return ret

    def itemChange(self, change, value):
        ret = self.itemChangeInner(change, value)
        if change in [
            self.GraphicsItemChange.ItemChildAddedChange,
            self.GraphicsItemChange.ItemChildRemovedChange,
        ]:
            try:
                itemsChangedListeners = self.itemsChangedListeners
            except AttributeError:
                # It's possible that the attribute was already collected when the itemChange happened
                # (if it was triggered during the gc of the object).
                pass
            else:
                for listener in itemsChangedListeners:
                    listener.itemsChanged()
        return ret
        
    def childItems(self):
        return list(self._GLGraphicsItem__children)

    def boundingRect(self):
        return QtCore.QRectF()
        
    def paint(self, *args):
        pass
    
    def addItem(self, item):
        item.setParentItem(self)
    
    def resetTransform(self):
        return super().resetTransform()

    

    def _internal_update(self):
        return QtWidgets.QGraphicsObject.update()

class AnotherItem(QtWidgets.QGraphicsItem):
    def __init__(self, parentItem=None):
        super().__init__(parent=parentItem)
        
    def _internal_update(self, **kwargs):...    
    
    

"""


class CustomItem(AbstractGLGraphicsItem, QtWidgets.QGraphicsObject):
    def __init__(self, parentItem=None):
        super().__init__(parentItem=parentItem)
        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        self._cachedView = None
        self.item_group = GLItemGroup(parentItem=self)
        self.setParentItem(parent)
    

"""

class CustomItem(QtWidgets.QGraphicsObject):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        
    def _internal_update(self, **kwargs):...    
    

    def boundingRect(self):
        return QtCore.QRectF(-10, -10, 20, 20)  # Or some non-zero rect
    
    def paint(self, *args, **kwargs):...


class QGraphicsWrapper(QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent=parent)     
        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        self.__depthValue=0

        self.custom_item = GLItemGroup()
        self.custom_item.setParentItem(self) 

    def _setView(self, view):
        self.__view=view
    
    def boundingRect(self):
        return QtCore.QRectF(-10, -10, 20, 20)  # Or some non-zero rect
    
    def paint(self, *args, **kwargs):...
    
    def isInitialized(self):
        return True
    
    def depthValue(self):
        """Return the depth value of this item. See setDepthValue for more information."""
        return self.__depthValue
    
    def visible(self):
        return self.isVisible()

#from pyvol_terminal.gl_3D_graphing.graphics_items.GLViewBox import GLViewBox

class GLViewBox(AbstractGLGraphicsItem, QtWidgets.QGraphicsItem):
    def __init__(self, parentItem=None):
        super().__init__()     
    
    def create_child(self):
        self.childGroup = GLItemGroup(parentItem=self)
        print(f"self.childGroup: {self.childGroup}")
        print(f'isQObjectAlive(self.childGroup): {isQObjectAlive(self.childGroup)}')
        if hasattr(type(self.childGroup), "implements"):
            print(True)
        else:
            print(False)
        
    def _internal_update(self, **kwargs):...    
    

    def boundingRect(self):
        return QtCore.QRectF(-10, -10, 20, 20)  # Or some non-zero rect
    
    def paint(self, *args, **kwargs):...


        


class CustomGLViewWidget(opengl.GLViewWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)     
#        wrapper = QGraphicsWrapper()
        self.vb=GLViewBox()
        self.addItem(self.vb)
        print("after addItem")
        self.vb.create_child()
        print(self.items)
        

class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)     
        
        #self.graphics_item_wrapper = QGraphicsWrapper()
        self.central_widget = CustomGLViewWidget()
        self.setCentralWidget(self.central_widget)
        self.central_layout = QtWidgets.QHBoxLayout(self.centralWidget())
        self.setLayout(self.central_layout)
        
        
        
        
        #self.central_widget.addItem(self.graphics_item_wrapper)



def main():
    app = QtWidgets.QApplication()
    win = Window()
    print(f"win.central_widget.items: {win.central_widget.items}")
    
    items = win.central_widget.items
    for item in items:
        
        print(f"item.childGroup: {item.childGroup}")
        
        parent_item = item.childGroup.parentItem()
        print(f"parent_item: {parent_item}")

        
        if hasattr(parent_item, "implements"):
            print(True)
        else:
            print(False)
        


            

    win.show()



    app.exec()

if __name__ == "__main__":
    main()


from PySide6 import QtWidgets
import sys
app = QtWidgets.QApplication(sys.argv)


s = QtWidgets.QSplitter()
w = QtWidgets.QWidget()
l = QtWidgets.QHBoxLayout()


print(issubclass(QtWidgets.QSplitter, QtWidgets.QWidget))

print(issubclass(QtWidgets.QSplitter, QtWidgets.QHBoxLayout))



import sys
import numpy as np
import pyqtgraph as pg
from PySide6.QtWidgets import (QApplication, QMainWindow, QWidget, 
                             QVBoxLayout, QHBoxLayout, QPushButton,
                             QStackedLayout, QLabel)
from PySide6.QtCore import Qt
from PySide6 import QtWidgets


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        
        self.setWindowTitle("Stacked Layout Example")
        self.setGeometry(100, 100, 800, 600)
        
        main_widget = QWidget()
        main_layout = QVBoxLayout()
        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)
        
        nav_layout = QHBoxLayout()
        self.btn1 = QPushButton("Graph View")
        self.btn2 = QPushButton("Text View")
        self.btn3 = QPushButton("Image View")
        
        self.btn1.clicked.connect(lambda: self.change_view(0))
        self.btn2.clicked.connect(lambda: self.change_view(1))
        self.btn3.clicked.connect(lambda: self.change_view(2))
        
        nav_layout.addWidget(self.btn1)
        nav_layout.addWidget(self.btn2)
        nav_layout.addWidget(self.btn3)
        main_layout.addLayout(nav_layout)
        

        self.stacked_layout = QStackedLayout()
        main_layout.addLayout(self.stacked_layout)
        
        self.create_graph_view()
        self.create_text_view()
        self.create_image_view()
        
        # Set the initial view
        self.change_view(0)
        
    def add_q_label(self):
        label = QtWidgets.QLabel(str(np.random.normal(0, 1, 1)))
        return label

    def create_graph_view(self):

        widget = QWidget()
        layout = QVBoxLayout()
        widget.setLayout(layout)  
        # Create plot widget
        plot_widget = pg.PlotWidget()
        
        settings = QHBoxLayout()
        labels = [self.add_q_label() for _ in range(4)]
        
        for label in labels:
            settings.addWidget(label)
        
        layout.addLayout(settings)
        layout.addWidget(plot_widget)
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        plot_widget.plot(x, y, pen='r')
        plot_widget.setTitle("Graph View - Click buttons above to change views")
        
        self.stacked_layout.addWidget(widget)  # Fixed this line
            
    def create_text_view(self):
        """Create the second view with text"""
        widget = QWidget()
        layout = QVBoxLayout()
        widget.setLayout(layout)
        
        label = QLabel("This is the Text View\n\n"
                      "You can put any text or widgets here.\n"
                      "Click the buttons above to switch between views.")
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(label)
        
        self.stacked_layout.addWidget(widget)
    
    def create_image_view(self):
        """Create the third view with an image"""
        widget = QWidget()
        layout = QVBoxLayout()
        widget.setLayout(layout)
        
        # Create a label for the title
        title_label = QLabel("Image View - Random Data")
        title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(title_label)
        
        # Create an image using PyQtGraph's ImageItem
        imv = pg.ImageView()
        
        # Create a sample image
        data = np.random.normal(size=(100, 100))
        data = np.cumsum(data, axis=0)
        data = np.sin(data * 0.1) * 10 + 10
        
        imv.setImage(data)
        layout.addWidget(imv)
        
        self.stacked_layout.addWidget(widget)
    
    def change_view(self, index):
        """Change the current view in the stacked layout"""
        self.stacked_layout.setCurrentIndex(index)
        
        # Update button styles to show which is active
        for i, btn in enumerate([self.btn1, self.btn2, self.btn3]):
            if i == index:
                btn.setStyleSheet("background-color: #4CAF50; color: white;")
            else:
                btn.setStyleSheet("")


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())


from __future__ import annotations
from typing import List, Dict, Union, Tuple, Callable, TYPE_CHECKING
if TYPE_CHECKING:
    from instruments.utils import InstrumentManager

from PySide6 import QtCore, QtWidgets, QtGui
from .. import windows
from pyvol_terminal import workers
from pyvol_terminal.data_classes import builders as builders_data_classes
from PySide6.QtMultimedia import QSoundEffect
class WarningClose(QtWidgets.QDialog):
    def __init__(self, geometry: QtCore.QRect, app_widget: Frontend, parent=None):
        super().__init__(parent=parent)
        self.app_widget = app_widget
        self._quitting = False
        
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(20, 20, 20, 20)  # Add some margins
        
        label = QtWidgets.QLabel("Close PyVol Surface?")
        label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        
        self.layout().addWidget(label)
        button_layout = QtWidgets.QHBoxLayout()
        
        self.button_yes = QtWidgets.QPushButton("Yes")
        self.button_no = QtWidgets.QPushButton("No")
        
        self.button_yes.clicked.connect(self._closeAppSlot)
        self.button_no.clicked.connect(self._keepAppOpenSlot)
        
        button_layout.addWidget(self.button_yes)
        button_layout.addWidget(self.button_no)
        self.layout().addLayout(button_layout)
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.hide()
    
    def show(self):
        parent_geometry = self.app_widget.geometry()
        half_width = parent_geometry.width() // 2
        
        self.adjustSize()
        natural_height = self.sizeHint().height()
        
        new_geometry = QtCore.QRect(parent_geometry.x() + (parent_geometry.width() - half_width) // 2,
                                    parent_geometry.y() + (parent_geometry.height() - natural_height) // 2,
                                    half_width,
                                    natural_height
                                    )
        
        self.setGeometry(new_geometry)
        super().show()
        
        self.setFixedSize(self.size())
    
    def _initLayout(self, geometry: QtCore.QRectF):
        self.setLayout(QtWidgets.QVBoxLayout())
        
        interface_center = geometry.center()
        geometry.setSize(geometry.size())
        geometry.moveCenter(interface_center)
        print(geometry)
            
        self.setGeometry(geometry)

    @QtCore.Slot()
    def _keepAppOpenSlot(self, *args, **kwargs):
        self.hide()
        self.app_widget._warned_user = False
        
    @QtCore.Slot()
    def _closeAppSlot(self, *args, **kwargs):
        QtWidgets.QApplication.instance().exit()
    
    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_Escape:
            self._keepAppOpenSlot()
        elif event.key() == QtCore.Qt.Key.Key_Shift:
            pass
        return super().keyPressEvent(event)
    
                
    def closeEvent(self, event):
        if self._quitting:
            super().closeEvent(event)
        else:
            event.ignore()
            self._keepAppOpenSlot()


class Frontend(QtWidgets.QMainWindow):
    def __init__(self,
                 open_window_signal: QtCore.Slot,
                 instruments: List[str],
                 interfaces: List[str],
                 parent=None
                 ) -> Frontend:
        super().__init__(parent)
        self.open_window_signal=open_window_signal
        self.instruments=instruments
        self.interfaces=interfaces
        
        self._warned_user=False
        self.warning_close_win = WarningClose(self.geometry(), app_widget=self)
        main_layout = QtWidgets.QVBoxLayout()
        
        self.buttons: List[QtWidgets.QPushButton] = []
        
        self.setCentralWidget(QtWidgets.QWidget())
        self.centralWidget().setLayout(main_layout)
        self._initLayoutContents()
        
        self.setMouseTracking(True)
        self.centralWidget().setMouseTracking(True)
        self._enableMouseTrackingForChildren(self.centralWidget())
            
    def _initLayoutContents(self):
        button_layout = QtWidgets.QHBoxLayout()
        
        for interface in self.interfaces:
            button = QtWidgets.QPushButton(interface)
            button.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed) 
            button.clicked.connect(self.create_button_handler(interface))
            button_width = 200  
            button_height = 100  
            button.setFixedSize(button_width, button_height)
            
            font = button.font()
            target_font_width = button_width // 1.5  
            target_font_height = button_height // 2  
            
            self.button_font_size(font, interface, button, target_font_width, target_font_height)
            
            self.buttons.append(button)
            button_layout.addWidget(button)
        
        self.combo_box = QtWidgets.QComboBox()
        self.combo_box.addItems(["No Option Chain Selected"] + self.instruments)
        self.combo_box.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        self.combo_box.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.combo_box.currentIndexChanged.connect(self.update_button_state)
        combo_font = self.combo_box.font()

        self.combo_box.setFont(combo_font)
        self.centralWidget().layout().addWidget(self.combo_box)
        self.centralWidget().layout().addLayout(button_layout)
        self.setStyleSheet("""
                           QPushButton { font-size: 28px; }
                           QComboBox { font-size: 14px; }
                           """
                           )
        
        self.update_button_state(0)
        self.adjustSize()              
        self.setFixedSize(self.size()) 
            
    def update_button_state(self, index):
        state = index != 0
        for button in self.buttons:
            button.setEnabled(state)
        if index == 0:
            for button in self.buttons:
                button.setStyleSheet("""
                                     QPushButton:disabled {background-color: #d0d0d0;  
                                     color: #666666;
                                     border: 1px solid #bbbbbb;}
                                     """
                                     )
        else:
            for button in self.buttons:
                button.setStyleSheet("")
        
    def closeEvent(self, event):
        return super().closeEvent(event)
        if self.warning_close_win.isVisible():
            event.ignore()
            self._warningCloseAttention()
            return 
        else:
            if not self._warned_user:
                event.ignore()
                self.warning_close_win.show()
                self.update()
                self._warned_user = True
            else:
                super().closeEvent(event)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_Escape:
            self.close()
        return super().keyPressEvent(event)
    
    def button_font_size(self,
                         font: QtGui.QFont,
                         button_string: str,
                         button_object: QtWidgets.QPushButton,
                         target_font_width,
                         target_font_height
                         ) -> None:
        font_size = 1
        while True:
            font.setPointSize(font_size)
            button_object.setFont(font)
            new_metrics = button_object.fontMetrics()
            if (new_metrics.horizontalAdvance(button_string) > target_font_width or 
                new_metrics.height() > target_font_height):
                font.setPointSize(font_size - 1) 
                button_object.setFont(font)
                break
            font_size += 1
    
    def create_button_handler(self, interface: str):
        captured_interface = interface
        @QtCore.Slot()
        def handler():
            selected_option = self.combo_box.currentText()
            return self.open_window_signal(selected_option, captured_interface)
        return handler

    def _enableMouseTrackingForChildren(self, widget):
        for child in widget.findChildren(QtWidgets.QWidget):
            child.setMouseTracking(True)
            self._enableMouseTrackingForChildren(child)

    def eventFilter(self, obj, event):
        if (event.type() in (QtCore.QEvent.MouseButtonPress, QtCore.QEvent.MouseButtonDblClick) and 
            self.warning_close_win.isVisible()):
            self._warningCloseAttention()
            return True
        return super().eventFilter(obj, event)

    def showEvent(self, event):
        super().showEvent(event)
        self.centralWidget().installEventFilter(self)
        for child in self.centralWidget().findChildren(QtWidgets.QWidget):
            child.installEventFilter(self)

    def _warningCloseAttention(self):
        self.warning_close_win.activateWindow()
        self.warning_close_win.setFocus()



import importlib

from OpenGL import GL
from OpenGL.GL import shaders
import numpy as np
from pyqtgraph.Qt import QT_LIB
from PySide6 import QtWidgets, QtCore, QtGui

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")



from pyqtgraph.Qt import QtGui, QT_LIB

from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
from .GL3DPlotDataItemMixin import GL3DFlatMeshPlotDataItemMixin, GL3DMeshPlotDataItemMixin

from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

from pyqtgraph.opengl.items.GLGraphItem import GLGraphItem
from pyqtgraph.opengl.items.GLGridItem import GLGridItem
from pyqtgraph.opengl.items.GLSurfacePlotItem import GLSurfacePlotItem
from pyqtgraph.opengl.items.GLBarGraphItem import GLBarGraphItem
from pyqtgraph.opengl.items.GLMeshItem import GLMeshItem
from pyqtgraph.opengl.items.GLLinePlotItem import GLLinePlotItem
from pyqtgraph.opengl.items.GLBoxItem import GLBoxItem
from pyqtgraph.opengl.items.GLVolumeItem import GLVolumeItem
from pyqtgraph.opengl.items.GLScatterPlotItem import GLScatterPlotItem
from pyqtgraph.opengl.items.GLTextItem import GLTextItem
from pyqtgraph.opengl.items.GLImageItem import GLImageItem
from pyqtgraph.opengl.items.GLAxisItem import GLAxisItem
from pyqtgraph.opengl.items.GLGradientLegendItem import GLGradientLegendItem
from pyqtgraph.opengl.MeshData import MeshData
import types



class GL3DMeshData(MeshData):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
    
    def vertexNormals(self, indexed=None):
        if self._vertexNormals is None:
            faceNorms = self.faceNormals()
            self._vertexNormals = np.zeros((self._vertexes.shape[0], 3), dtype=np.float32)
            
            # Vectorized scattering of face normals to vertices
            for i in range(3):  # Process each vertex in the triangle
                np.add.at(self._vertexNormals, self._faces[:, i], faceNorms)
            
            # Vectorized normalization
            mag = np.linalg.norm(self._vertexNormals, axis=1)
            non_zero_mask = mag > 0
            self._vertexNormals[non_zero_mask] /= mag[non_zero_mask, np.newaxis]

        if indexed is None:
            return self._vertexNormals
        elif indexed == 'faces':
            return self._vertexNormals[self.faces()]
        else:
            raise Exception("Invalid indexing mode. Accepts: None, 'faces'")




class GL3DMeshItem(GL3DGraphicsItemMixin, GLMeshItem):
    def __init__(self, parentItem=None, **kwds):
        super().__init__(parentItem=parentItem, **kwds)

        
        
        glopts = kwds.pop('glOptions', 'opaque')
        self.setGLOptions(glopts)
        shader = kwds.pop('shader', None)
        self.setShader(shader)
        
        self.setMeshData(**kwds)
        
        ## storage for data compiled from MeshData object
        self.vertexes = None
        self.normals = None
        self.colors = None
        self.faces = None

        self.m_vbo_position = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_normal = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_color = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_ibo_faces = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.IndexBuffer)
        self.m_vbo_edgeVerts = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_ibo_edges = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.IndexBuffer)
    
    def _childPaint(self): GLMeshItem.paint()

    def _paintHelper(self): super()._paintHelper()

    def setMeshData(self, **kwds):
        md = kwds.get('meshdata', None)
        if md is None:
            opts = {}
            for k in ['vertexes', 'faces', 'edges', 'vertexColors', 'faceColors']:
                try:
                    opts[k] = kwds.pop(k)
                except KeyError:
                    pass
            md = GL3DMeshData(**opts)
        
        self.opts['meshdata'] = md
        self.opts.update(kwds)
        self.meshDataChanged()
        self.update()
    



class GL3DLinePlotDataItem(GL3DFlatMeshPlotDataItemMixin, GLLinePlotItem):
    _shaderProgram = None

    def __init__(self, parentItem=None, **kwds):
        """All keyword arguments are passed to setData()"""
        super().__init__(**kwds)
        glopts = kwds.pop('glOptions', 'additive')
        self.setGLOptions(glopts)
        self.pos = None
        self.mode = 'lines'
        self.width = 1.
        self.color = (1.0,1.0,1.0,1.0)
        self.antialias = False

        self.m_vbo_position = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_color = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.vbos_uploaded = False
        self.setParentItem(parentItem)
        self.setData(**kwds)

    @property
    def dataAttr(self):
        return self.pos

    def _childPaint(self):
        return GLLinePlotItem.paint(self)
    
    def _setDataChild(self, *args, **kwds):
        return GLLinePlotItem.setData(self, *args, **kwds)
    
    
class MixedGLLinePlotItem(GL3DGraphicsItemMixin, GLLinePlotItem):
    _shaderProgram = None

    def __init__(self, parentItem=None, **kwds):
        """All keyword arguments are passed to setData()"""
        super().__init__(**kwds)
        glopts = kwds.pop('glOptions', 'additive')
        self.setGLOptions(glopts)
        self.pos = None
        self.mode = 'line_strip'
        self.width = 1.
        self.color = (1.0,1.0,1.0,1.0)
        self.antialias = False

        self.m_vbo_position = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_color = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.vbos_uploaded = False
        self.setParentItem(parentItem)
        self.setData(**kwds)

    def _paintHelper(self):
        return super()._paintHelper()
    
    def _childPaint(self):
        return GLLinePlotItem.paint(self)

    
    def childrenBoundingRect(self):...
    
        
class MixedGLGraphItem(GL3DGraphicsItemMixin, GLGraphicsItem):
    def __init__(self, parentItem=None, **kwds):
        super().__init__(parentItem=parentItem)
        glopts = kwds.pop('glOptions', 'translucent')

        self.edges = None
        self.edgeColor = QtGui.QColor(QtCore.Qt.GlobalColor.white)
        self.edgeWidth = 1.0

        self.lineplot = GL3DLinePlotDataItem(parentItem=self, glOptions=glopts, mode='lines')
        self.scatter = GL3DLinePlotDataItem(parentItem=self, glOptions=glopts)
        self.setParentItem(parentItem)
        self.setData(**kwds)

class MixedGLGridItem(GL3DGraphicsItemMixin, GLGridItem):
    def __init__(self, size=None, color=(255, 255, 255, 76.5), antialias=True, glOptions='translucent', parentItem=None):
        super().__init__(parentItem=parentItem)

        self.lineplot = None    # mark that we are still initializing

        if size is None:
            size = QtGui.QVector3D(20,20,1)
        self.setSize(size=size)
        self.setSpacing(1, 1, 1)
        self.setColor(color)

        self.lineplot = GL3DLinePlotDataItem(parentItem=self, glOptions=glOptions, mode='lines', antialias=antialias)
        
        self.setParentItem(parentItem)
        self.updateLines()


def GL3DvertexNormals(self, indexed=None):
    if self._vertexNormals is None:
        faceNorms = self.faceNormals()
        self._vertexNormals = np.zeros((self._vertexes.shape[0], 3), dtype=np.float32)
        
        # Vectorized scattering of face normals to vertices
        for i in range(3):  # Process each vertex in the triangle
            np.add.at(self._vertexNormals, self._faces[:, i], faceNorms)
        
        # Vectorized normalization
        mag = np.linalg.norm(self._vertexNormals, axis=1)
        non_zero_mask = mag > 0
        self._vertexNormals[non_zero_mask] /= mag[non_zero_mask, np.newaxis]

    if indexed is None:
        return self._vertexNormals
    elif indexed == 'faces':
        return self._vertexNormals[self.faces()]
    else:
        raise Exception("Invalid indexing mode. Accepts: None, 'faces'")



    

class GL3DSurfacePlotItem(GL3DMeshPlotDataItemMixin, GLSurfacePlotItem):
    def __init__(self, x=None, y=None, z=None, colors=None, parentItem=None, **kwds):
        
        self._GL3DMeshData=None
        
        super().__init__(x=x, y=y, z=z, colors=colors, parentItem=parentItem, **kwds)
    
    @property
    def _meshdata(self):
        return self._GL3DMeshData
    
    @_meshdata.setter
    def _meshdata(self, value):
        
        if not hasattr(value, "GL3DOptimization"):
            value.vertexNormals = types.MethodType(GL3DvertexNormals, value)
            setattr(value, "GL3DOptimization", True)
        
        self._GL3DMeshData = value
    
    
    @property
    def dataAttr(self):
        return self._x, self._y, self._z

    def _childPaint(self):
        return GLSurfacePlotItem.paint(self)
    
    def _setDataChild(self, *args, **kwds):
        return GLSurfacePlotItem.setData(self, *args, **kwds)
    
    def setMeshData(self, **kwds):
        """
        Set mesh data for this item. This can be invoked two ways:
        
        1. Specify *meshdata* argument with a new MeshData object
        2. Specify keyword arguments to be passed to MeshData(..) to create a new instance.
        """
        md = kwds.get('meshdata', None)
        if md is None:
            opts = {}
            for k in ['vertexes', 'faces', 'edges', 'vertexColors', 'faceColors']:
                try:
                    opts[k] = kwds.pop(k)
                except KeyError:
                    pass
            md = MeshData(**opts)
        
        if not hasattr(md, "GL3DOptimization"):
            md.vertexNormals = types.MethodType(GL3DvertexNormals, md)
            setattr(md, "GL3DOptimization", True)
        
        self.opts['meshdata'] = md
        self.opts.update(kwds)
        self.meshDataChanged()
        self.update()
        
    

        
        
class MixedGLBarGraphItem(GL3DGraphicsItemMixin, GLBarGraphItem):
    def __init__(self, pos, size, parentItem=None):
        super().__init__(pos, size, parentItem=parentItem)


class MixedGLBoxItem(GL3DGraphicsItemMixin, GLBoxItem):
    def __init__(self, size=None, color=None, glOptions='translucent', parentItem=None):
        super().__init__(size=size, color=color, glOptions=glOptions, parentItem=parentItem)

        self.lineplot = None 

        if size is None:
            size = QtGui.QVector3D(1,1,1)
        self.setSize(size=size)
        if color is None:
            color = (255,255,255,80)
        self.setColor(color)

        self.lineplot = GL3DLinePlotDataItem(parentItem=self, glOptions=glOptions, mode='lines')
        self.setParentItem(parentItem)
        self.updateLines()

class MixedGLVolumeItem(GLVolumeItem):
    _shaderProgram = None
    
    def __init__(self, data, sliceDensity=1, smooth=True, glOptions='translucent', parentItem=None):
        super().__init__(data, sliceDensity=sliceDensity, smooth=smooth, glOptions=glOptions, parentItem=parentItem)


class GL3DScatterPlotItem(GL3DFlatMeshPlotDataItemMixin, GLScatterPlotItem):
    def __init__(self, parentItem=None, **kwds):
        super().__init__(parentItem=parentItem, **kwds)
        
    @property
    def dataAttr(self):
        return self.pos

    def _childPaint(self):
        return GLScatterPlotItem.paint(self)
    
    def _setDataChild(self, *args, **kwds):
        return GLScatterPlotItem.setData(self, *args, **kwds)


class MixedGL3DTextItem(GL3DGraphicsItemMixin, GLTextItem):
    def __init__(self, parentItem=None, **kwds):
        super().__init__(parentItem=parentItem, **kwds)
        
    def _childPaint(self):return GL3DGraphicsItemMixin.paint(self)


class MixedGLImageItem(GL3DGraphicsItemMixin, GLImageItem):
    def __init__(self, data, smooth=False, glOptions='translucent', parentItem=None):
        super().__init__(data, smooth=smooth, glOptions=glOptions, parentItem=parentItem)


class MixedGLAxisItem(GL3DGraphicsItemMixin, GLAxisItem):
    def __init__(self, size=None, antialias=True, glOptions='translucent', parentItem=None):
        super().__init__(size=size, antialias=antialias, glOptions=glOptions, parentItem=parentItem)
        self.lineplot = None    # mark that we are still initializing

        if size is None:
            size = QtGui.QVector3D(1,1,1)
        self.setSize(size=size)

        self.lineplot = GL3DLinePlotDataItem(parentItem=self, glOptions=glOptions, mode='lines', antialias=antialias)
        self.setParentItem(parentItem)
        self.updateLines()

class MixedGLGradientLegendItem(GL3DGraphicsItemMixin, GLGradientLegendItem):
    def __init__(self, parentItem=None, **kwds):
        super().__init__(parentItem=parentItem, **kwds)


from __future__ import annotations 
from typing import List, Optional, Any, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option, ABCInstrument, Future, Spot, ABCNestedInstrument
    from pandas import DataFrame
import numpy as np
import pandas as pd
from dataclasses import dataclass, field
from collections import defaultdict

    
@dataclass(slots=True, frozen=True)
class BaseMap:
    index_name_map: Dict
    name_index_map: Dict
    
    
@dataclass(slots=True, frozen=True, kw_only=True)
class OptionMap(BaseMap):
    put_call_map: Dict[str, str] = field(default_factory=dict)
    underlying_ticker_map: Dict[str, List[str]] = field(default_factory=dict)
    name_underlying_map: Dict[str, str] = field(default_factory=dict)
    expiry_strike_map: Dict[float, np.ndarray] = field(default_factory=dict)
    strike_expiry_map: Dict[float, np.ndarray] = field(default_factory=dict)
    expiry_instrument_map: Dict[str, List[str]] = field(default_factory=dict)
    strike_instrument_map: Dict[float, List[str]] = field(default_factory=dict)
    expiry_strike_instrument_map: Dict[str, Dict[float, List[str]]] = field(default_factory=dict)
    strike_expiry_instrument_map: Dict[float, Dict[str, List[str]]] = field(default_factory=dict)
    expiry_strike_type_instrument_map: Dict[float, Dict[float, Dict[str, str]]] = field(default_factory=dict)
    strike_expiry_type_instrument_map: Dict[float, Dict[str, Dict[str, str]]] = field(default_factory=dict)
    type_expiry_strike_instrument_map: Dict[str, Dict[float, Dict[float, str]]] = field(default_factory=dict)

@dataclass(slots=True, frozen=True)
class ManagerMap:
    name_type_map: Dict[str, str] = field(default_factory=dict)

@dataclass(slots=True)
class InstrumentContainer:
    objects: Dict[str, ABCInstrument]
    price_types: List[str,]
    maps: OptionMap
    instrument_type: str
    original_data: pd.DataFrame
    
    def get_objects(self) -> List[ABCInstrument]:
        return list(self.objects.values())
    
@dataclass(slots=True)
class NestedInstrumentContainer:
    objects: Dict[str, ABCNestedInstrument]
    price_types: List[str,]
    maps: Dict[str, Any]
    instrument_type: str
    original_data: pd.DataFrame
    
    def get_objects(self) -> List[ABCInstrument]:
        return self.get_nested_object()
    
    def get_nested_object(self):
        total_objects = []
        for obj in self.objects.values():
            total_objects.append(obj.get_nested_objects())
        return total_objects
    

    

@dataclass(slots=True, kw_only=True)
class InstrumentManager:
    config: Dict[str, Dict[str, Any]]
    name: str = field(default=None)
    manager_map: ManagerMap
    all_instrument_objects: Dict[str, ABCInstrument] = field(default_factory=dict)
    spot_instrument_container: InstrumentContainer = field(default=None)
    futures_instrument_container: InstrumentContainer = field(default=None)
    options_instrument_container: InstrumentContainer = field(default=None)
    
    def __post_init__(self):
        if self.name is None:
            if not self.spot_instrument_container is None:
                self.name = [obj.ticker for obj in self.spot_instrument_container.get_objects()][0]
            if self.name is None and not self.futures_instrument_container is None:
                self.name = longest_common_prefix([obj.ticker for obj in self.futures_instrument_container.get_objects()])
            if self.name is None and not self.options_instrument_container is None:
                self.name = longest_common_prefix([obj.ticker for obj in self.options_instrument_container.get_objects()])

    def update_metric_calculations(self,
                                   flag: bool,
                                   option_ptype: str,
                                   underyling_ptype: str
                                   ):
        for option in self.options_instrument_container.get_objects():
            if flag:
                option.add_active_ivol_ptype(metric_calculation)
            else:
                option.remove_active_ivol_ptype(metric_calculation)
    

def longest_common_prefix(strings: List[str]):
    prefix = strings[0]
    for s in strings[1:]:
        while not s.startswith(prefix):
            prefix = prefix[:-1]
            if not prefix:
                return ""
    return prefix


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option, Spot
    from ..graphics_items.GL3DGraphicsItems import GlPlotItemMixin
    from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

from PySide6 import QtCore, QtGui
import numpy as np
import numpy as np
from dataclasses import dataclass, InitVar, field
from PySide6 import QtCore
from ..graphics_items import GL3DAxisItem
from .. import utils
from pprint import pprint   

@dataclass(slots=True)
class PlotView:
    axis: str
    axis_idx: int
    min: float = np.nan
    max: float = np.nan
    min_UB: float = np.nan
    max_LB: float = np.nan
    
    def normalise(self, values):
        return (values - self.min) / (self.max - self.min)
    
    def unnormalise(self, values):
        return values * (self.max - self.min) + self.min
    
    def update_min(self, min):
        self.min = min
        self._get_min_UB()
    
    def update_max(self, max):
        self.max = max
        self._get_max_LB()
    
    def _get_min_UB(self):
        if not np.isnan(self.max):
            d = self.max - self.min
            self.min_UB = self.min + 0.4 * d
            
    def _get_max_LB(self):
        if not np.isnan(self.min):
            d = self.max - self.min
            self.max_LB = self.max - 0.4 * d

    def is_new_limit(self, limit_type, value):
        if np.isnan(getattr(self, limit_type)):
            return True
        else:
            if limit_type == "min":
                if np.isnan(self.min_UB) and not np.isnan(self.max):
                    self._get_min_UB()
                if value < self.min or value > self.min_UB:
                    return True
                else:
                    return False
            else:
                if np.isnan(self.max_LB) and not np.isnan(self.min):
                    self._get_max_LB()
                if value > self.max or value < self.max_LB:
                    return True
                else:
                    return False
    
    def mask_from_data(self, data):
        return (data >= self.min) & (data <= self.max)
        
        
class ViewBox(QtCore.QObject):
    sigXRangeChanged = QtCore.Signal(PlotView, QtCore.QObject)
    sigYRangeChanged = QtCore.Signal(PlotView, QtCore.QObject)
    sigZRangeChanged = QtCore.Signal(PlotView, QtCore.QObject)

    sigXNormChanged = QtCore.Signal(PlotView)
    sigYNormChanged = QtCore.Signal(PlotView)
    sigZNormChanged = QtCore.Signal(PlotView)

    def __init__(self, norm_dims=None, padding=None):
        super().__init__()
        self.norm_dims=norm_dims
        self.item_limit_engine = utils.ItemSetLimits("xyz")
        self.require_normalise=False
        if norm_dims is None:
            self.norm_dims = {ax : 1 for ax in range(3)}
        else:
            self.norm_dims = norm_dims

        if padding is None:
            self.padding = {ax : 0 for ax in range(3)}
        else:
            self.padding = padding
        self._item_limits={}
        self._item_internal_normalise_name_map={}
        self.overall_item_limits=self._get_null_limits()
        self._normalise_flags={}
        self.plot_views = {axis : PlotView(axis, idx) for idx, axis in enumerate(range(3))}
        self.viewRange = [[0, 0] for _ in range(3)]
        self.tt=[]
        self.addedItems=[]
        
    
    def addItem(self,
                item: GLGraphicsItem,
                ignoreBounds: bool=False
                ):
        if hasattr(item, "type"):
            if item.type=="axis":
                return
        if isinstance(item, GL3DAxisItem.GL3DViewBox):
            getattr(self, f"sig{item.axis_direction.upper()}RangeChanged").connect(item.update_values)
        else:
            if not ignoreBounds and hasattr(item, "setData") and callable(item.setData):
                original_setData = item.setData
                item.original_setData=original_setData
                
                self._normalise_flags[item.id()]=False 
                if not ignoreBounds:
                    self._add_item_check_limits(item)
                
          
                    
                def _normalise_setData_pos_wrapper(*inner_args, **inner_kwargs):
                    return utils.normalise_setData_pos(item, self, ignoreBounds, *inner_args, **inner_kwargs)
                
                def _renormalise_data_pos_wrapper(plot_view):
                    return utils.renormalise_data_pos(item, plot_view)
                
                def _filter_data_pos_wrapper(plot_view):
                    return utils.filter_data_pos(item, plot_view)

                def _normalise_setData_xyz_wrapper(*inner_args, **inner_kwargs):
                    return utils.normalise_setData_xyz(item, self, ignoreBounds, *inner_args, **inner_kwargs)
                
                def _renormalise_data_xyz_wrapper(plot_view):
                    return utils.renormalise_data_xyz(item, plot_view)
                
                def _filter_data_xyz_wrapper(plot_view):
                    return utils.filter_data_xyz(item, plot_view, self)
                
                if hasattr(item, "pos"):
                    item.setData = _normalise_setData_pos_wrapper
               #     item.renormalise_data = _renormalise_data_pos_wrapper
                   # item.filter_data = _filter_data_pos_wrapper
                    item.unnorm_pos=item.pos.copy()
                    setData_kwargs = {"pos" : item.pos}

                elif hasattr(item, "_x") and hasattr(item, "_y") and hasattr(item, "_z"):
                  #  item.unnorm_x, item.unnorm_y, item.unnorm_z = item._x.copy(), item._y.copy(), item._z.copy()
                #    item._all_x, item._all_y, item._all_z = item._x.copy(), item._y.copy(), item._z.copy()
              #      item.unnorm_pos = np.column_stack((item.unnorm_x, item.unnorm_y, item.unnorm_z))
                    item.setData = _normalise_setData_xyz_wrapper
               #     item.renormalise_data = _renormalise_data_xyz_wrapper   
                  #  item.filter_data = _filter_data_xyz_wrapper
                    setData_kwargs = {"x":item._x, "y":item._y, "z":item._z}
                
                self.blockSignals(True)
                self.connect_slot_to_norm_sig(item, item.renormalise_data)
                self.connect_slot_to_rangechanged_sig(item, item.filter_data)
                self.blockSignals(False)
                item.setData(**setData_kwargs)
                self.addedItems.append(item)
        return item

    def removeItem(self, item):
        if item in self.addedItems:
            self.addedItems.remove(item)
            
    
    def reset_normalisation(self):
        self.plot_views = {axis : PlotView(axis) for axis in range(3)}
        self.item_limit_engine = utils.ItemSetLimits("xyz")
        
    def update_normalisation(self, item, pos=None, x=None, y=None, z=None):
        if not pos is None:
            x=pos[:,0]
            y=pos[:,1]
            z=pos[:,2]
        
        for axis, arr in zip(range(3), (x,y,z)):
            self.item_limit_engine.update_item_limits(item, axis, arr)
            self._check_view_box_limits(axis, item.id(), ignore_update_signals=True)
    
    def update_plot_view(self, plot_view):
        self.viewRange[plot_view.axis_idx][0] = plot_view.min
        self.viewRange[plot_view.axis_idx][1] = plot_view.max

        getattr(self, f"sig{plot_view.axis.upper()}RangeChanged").emit(plot_view.axis, self)

    def _check_view_box_limits(self, ax, item_id, ignore_update_signals=False):
        print(f"\n_check_view_box_limits")
        limits=self.item_limit_engine.get_limit_values(ax)
        plot_view=self.plot_views[ax]
        for limit, limit_type in zip(limits, ["min", "max"]):
            if limit != np.inf:
                if plot_view.is_new_limit(limit_type, limit):
                    self.require_normalise=True
                    padding = - self.padding[ax] if limit_type == "min" else self.padding[ax]
                    getattr(plot_view, f"update_{limit_type}")(limit + padding*limit)
                    
        self.viewRange[plot_view.axis_idx][0] = plot_view.min
        self.viewRange[plot_view.axis_idx][1] = plot_view.max
        pprint(self.viewRange)
        if self.require_normalise and not ignore_update_signals:
            self.require_normalise=False
            self._normalise_flags[item_id]=True
            #getattr(self, f"sig{ax.upper()}RangeChanged").emit(plot_view.axis, self)
            getattr(self, f"sig{ax.upper()}NormChanged").emit(plot_view)
            self._normalise_flags[item_id]=False
    
    def _add_item_check_limits(self, item):
        self.item_limit_engine.addItem(item)
        for axis in range(3):
            self._check_view_box_limits(axis, item.id())
        
    def _pre_normalisation(self, item: GlPlotItemMixin, axis, values):
        self.item_limit_engine.update_item_limits(item, axis, values)
        self._check_view_box_limits(axis, item.id())
            
    def normalise(self, axis, item, values, ignoreBounds=False):
        if not ignoreBounds:
            if not self._normalise_flags[item.id()]:
                self._pre_normalisation(item, axis, values)
        return self.plot_views[axis].normalise(values)

    def connect_slot_to_norm_sig(self, internal_name, slot):
        self._normalise_flags[internal_name]=False
        self.blockSignals(True)
        for ax in "XYZ":
            getattr(self, f"sig{ax}NormChanged").connect(slot)
        self.blockSignals(False)
        
    def connect_slot_to_rangechanged_sig(self, internal_name, slot):
        self._normalise_flags[internal_name]=False
        self.blockSignals(True)
        for ax in "XYZ":
            getattr(self, f"sig{ax}RangeChanged").connect(slot)
        self.blockSignals(False)

    @classmethod
    def _get_null_limits(cls):
        return {"x" : [np.nan, np.nan], "y" : [np.nan, np.nan], "z" : [np.nan, np.nan]} 

    def setRange(self, xRange=None, yRange=None, zRange=None, padding=None):
        if all((xRange is None,
                yRange is None,
                zRange is None)):
            for axis in "xyz":
                min_val, max_val = self.item_limit_engine.get_limit_values(axis)
                getattr(self, f"sig{axis.upper()}RangeChanged").emit(
                    PlotView(axis, list("xyz").index(axis), min_val, max_val), 
                    self
                )
            return

        if not xRange is None:
            plot_view=self.plot_views["x"]
            plot_view.update_min(xRange[0])
            plot_view.update_max(xRange[1])
            self.update_plot_view(plot_view)

        if not yRange is None:
            plot_view=self.plot_views["y"]
            plot_view.update_min(yRange[0])
            plot_view.update_max(yRange[1])
            self.update_plot_view(plot_view)
            
        if not zRange is None:
            plot_view=self.plot_views["z"]
            plot_view.update_min(zRange[0])
            plot_view.update_max(zRange[1])
            self.update_plot_view(plot_view)
        
        for item in self.addedItems:
            for axes in range(3):
                self.item_limit_engine.update_item_limits(item,
                                                          axes,
                                                          )
                
    def resetToDataLimits(self):
        for axis in range(3):
            min_val, max_val = self.item_limit_engine.get_limit_values(axis)
            self.setRange(**{f"{axis}Range": [min_val, max_val]})


"""
This example demonstrates the use of GLSurfacePlotItem.
"""
import sys

import numpy as np

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui
import pyqtgraph.opengl as gl
from pyqtgraph.Qt import QtCore

if 'darwin' in sys.platform:
    fmt = QtGui.QSurfaceFormat()
    fmt.setRenderableType(fmt.RenderableType.OpenGL)
    fmt.setProfile(fmt.OpenGLContextProfile.CoreProfile)
    fmt.setVersion(4, 1)
    QtGui.QSurfaceFormat.setDefaultFormat(fmt)

## Create a GL View widget to display data
app = pg.mkQApp("GLSurfacePlot Example")
w = gl.GLViewWidget()
w.show()
w.setWindowTitle('pyqtgraph example: GLSurfacePlot')
w.setCameraPosition(distance=50)

## Add a grid to the view
g = gl.GLGridItem()
g.scale(2,2,1)
g.setDepthValue(10)  # draw grid after surfaces since they may be translucent
w.addItem(g)


## Simple surface plot example
## x, y values are not specified, so assumed to be 0:50
z = pg.gaussianFilter(np.random.normal(size=(50,50)), (1,1))
p1 = gl.GLSurfacePlotItem(z=z, shader='shaded', color=(0.5, 0.5, 1, 1))
p1.scale(16./49., 16./49., 1.0)
p1.translate(-18, 2, 0)
w.addItem(p1)


## Saddle example with x and y specified
x = np.linspace(-8, 8, 50)
y = np.linspace(-8, 8, 50)
z = 0.1 * ((x.reshape(50,1) ** 2) - (y.reshape(1,50) ** 2))
p2 = gl.GLSurfacePlotItem(x=x, y=y, z=z, shader='normalColor')
p2.translate(-10,-10,0)
w.addItem(p2)


## Manually specified colors
z = pg.gaussianFilter(np.random.normal(size=(50,50)), (1,1))
x = np.linspace(-12, 12, 50)
y = np.linspace(-12, 12, 50)
colors = np.ones((50,50,4), dtype=np.float32)
colors[...,0] = np.clip(np.cos(((x.reshape(50,1) ** 2) + (y.reshape(1,50) ** 2)) ** 0.5), 0, 1)
colors[...,1] = colors[...,0]

p3 = gl.GLSurfacePlotItem(z=z, colors=colors.reshape(50*50,4), shader='shaded', smooth=False)
p3.scale(16./49., 16./49., 1.0)
p3.translate(2, -18, 0)
w.addItem(p3)




## Animated example
## compute surface vertex data
cols = 90
rows = 100
x = np.linspace(-8, 8, cols+1).reshape(cols+1,1)
y = np.linspace(-8, 8, rows+1).reshape(1,rows+1)
d = (x**2 + y**2) * 0.1
d2 = d ** 0.5 + 0.1

## precompute height values for all frames
phi = np.arange(0, np.pi*2, np.pi/20.)
z = np.sin(d[np.newaxis,...] + phi.reshape(phi.shape[0], 1, 1)) / d2[np.newaxis,...]


## create a surface plot, tell it to use the 'heightColor' shader
## since this does not require normal vectors to render (thus we 
## can set computeNormals=False to save time when the mesh updates)
p4 = gl.GLSurfacePlotItem(x=x[:,0], y = y[0,:], shader='heightColor', computeNormals=False, smooth=False)
p4.shader()['colorMap'] = np.array([0.2, 2, 0.5, 0.2, 1, 1, 0.2, 0, 2])
p4.translate(10, 10, 0)
w.addItem(p4)

index = 0
def update():
    global p4, z, index
    index -= 1
    p4.setData(z=z[index%z.shape[0]])
    
timer = QtCore.QTimer()
timer.timeout.connect(update)
timer.start(30)

if __name__ == '__main__':
    pg.exec()



#%% 

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
import math

class CubeVertex:
    def __init__(self, axis, point, perp_axis_edge, adj_axis_edge, perpendicular_size, adjacent_size):
        self.axis=axis
        self.point = point
        self.perp_axis_edge=perp_axis_edge
        self.adj_axis_edge = adj_axis_edge
        self.perpendicular_size = perpendicular_size
        self.adjacent_size = adjacent_size

        self.slope = self._slope()
      #  self.
    
    def _slope(self):
        dx = self.adj_axis_edge[0] - self.point[0]
        dy = self.adj_axis_edge[1] - self.point[1]
        slope = dy/dx if dx != 0 else float('inf')
        return round(slope, 1)


def anticlockwise_angle(p1, p2, k2):
    v1 = np.array(p2) - np.array(p1)
    v2 = np.array(k2) - np.array(p1)

    dot = np.dot(v1, v2)
    det = v1[0]*v2[1] - v1[1]*v2[0]  # 2D cross product

    angle_rad = np.arctan2(det, dot)  # signed angle (-π to π)
    anticlockwise_angle = angle_rad % (2 * np.pi)  # wrap to [0, 2π)
    return np.degrees(anticlockwise_angle)


def clockwise_angle(p1, p2, k2):
    v1 = np.array(p2) - np.array(p1)
    v2 = np.array(k2) - np.array(p1)

    dot = np.dot(v1, v2)
    det = v1[0]*v2[1] - v1[1]*v2[0]  

    angle_rad = np.arctan2(det, dot) 
    clockwise_angle = (-angle_rad) % (2 * np.pi)
    return np.degrees(clockwise_angle)

def offset_from_p1_viewed_from_p2(p1, p2, distance=5, side="left"):
    p1 = np.array(p1)
    p2 = np.array(p2)
    
    # Direction vector from p2 to p1
    direction = p1 - p2
    
    # Normalize direction vector
    norm = np.linalg.norm(direction)
    if norm == 0:
        raise ValueError("p1 and p2 cannot be the same point.")
    direction = direction / norm
    
    # Perpendicular direction
    if side == "left":
        perp = np.array([-direction[1], direction[0]])  # 90° CCW
    elif side == "right":
        perp = np.array([direction[1], -direction[0]])  # 90° CW
    else:
        raise ValueError("side must be 'left' or 'right'")
    
    return p1 + distance * perp  


def get_boundary_intersections(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    intersections = []
    
    if x1 == x2:
        return [(x1, -10), (x1, 10)]
    
    m = (y2 - y1)/(x2 - x1)
    b = y1 - m*x1
    
    y = m*(-10) + b
    if -10 <= y <= 10:
        intersections.append((-10, y))
        
    y = m*10 + b
    if -10 <= y <= 10:
        intersections.append((10, y))
        
    x = (-10 - b)/m
    if -10 <= x <= 10:
        intersections.append((x, -10))
        
    x = (10 - b)/m
    if -10 <= x <= 10:
        intersections.append((x, 10))
        
    return intersections

def intersection_from_point_slope(p1, m1, p2, m2):
    x1, y1 = p1
    x2, y2 = p2

    if m1 == m2:
        return None  # Lines are parallel, no intersection (or infinite if same line)

    # Handle vertical lines (infinite slope)
    if np.isinf(m1):
        x = x1
        y = m2 * (x - x2) + y2
    elif np.isinf(m2):
        x = x2
        y = m1 * (x - x1) + y1
    else:
        x = (m1 * x1 - m2 * x2 + y2 - y1) / (m1 - m2)
        y = m1 * (x - x1) + y1

    return x, y

def left_or_right(p1, p2, p3):
    p1 = np.array(p1)
    p2 = np.array(p2)
    p3 = np.array(p3)

    v1 = p2 - p1
    v2 = p3 - p1

    cross = v1[0] * v2[1] - v1[1] * v2[0]

    if cross > 0:
        return 'left'
    elif cross < 0:
        return 'right'
    else:
        return 'on the line'
    
    
def plot_vertex_lines(ax: Axes, p2: float, m: float, length: float = 0.3):
    cube_vertices = get_cube_coordinates()
    cube_vertices = [*cube_vertices[:3], cube_vertices[3]]
    degrees = []
    degrees_anti = []
    idx=0
    for vertex in cube_vertices:
      #  if vertex.point == (0, 1):
       #     continue
        x0, y0 = vertex.point
        if np.isinf(m):
            dx, dy = 0, length / 2
        else:
            dx = (length / 2) / np.sqrt(1 + m**2)
            dy = m * dx
            
        m1 = - 1 / m   # vertex slope
        m2 = m
        
        point_intersection = intersection_from_point_slope(vertex.point, m1, p2, m2)

            
        x_vals = [x0 - dx, x0 + dx]
        y_vals = [y0 - dy, y0 + dy]
        
        x1, y1 = offset_from_p1_viewed_from_p2(vertex.point, point_intersection, distance=length, side="left")
        axis_tick_origin = offset_from_p1_viewed_from_p2(vertex.adj_axis_edge, point_intersection, distance=length, side="left")
        
        axis_tick_pos = [(0, 0) for _ in range(5)]
      #  for idx, size in :
     #       axis_tick_pos[vertex.axis] = size
     
        start_tick_pos = [(0, 0) for _ in range(5)]


        ax_input = [ np.zeros(5).tolist(), np.zeros(5).tolist()]        

        ax_input[1 - vertex.axis] = 0.3
        end_tick_pos = []
        
        increments = (np.array(vertex.adj_axis_edge) - np.array(vertex.point)) / 5
        pos_s = list(vertex.point)
        for idx in range(6):
            
            pos_e = pos_s.copy()
            pos_e[1 - vertex.axis] += 0.3
            
            
            tick_x = [pos_s[0], pos_e[0]]
            tick_y = [pos_s[1], pos_e[1]]
            
            ax.plot(tick_x, tick_y, color='red')
            
            pos_s += increments
        print(increments)
        
        side = left_or_right(vertex.point, vertex.adj_axis_edge, p2)
            

            
        
        

        x_vals = [x0 , x1]
        y_vals = [y0 , y1]
        
        
        
        
      #  ax.plot(x_vals, y_vals, color='orange')
        
        
        
        ax.text(x0, y0,
                str(vertex.adj_axis_edge) ,#+ f"\n{idx}",
                #side,
                ha='center', va='center', color="k", size=7
                )
        
        
        deg = clockwise_angle(vertex.point, vertex.adj_axis_edge, p2)
        deg_rounded = deg % 90
        
        point_intersection_rounded = np.round(point_intersection, 1).tolist()
        
        
        string = str(round(deg, 1)) + f"     ({str(round(deg_rounded, 1))})"#f"     {str(tuple(point_intersection_rounded))}"
        degrees.append(string)
        idx+=1
    
    return degrees


def get_cube_coordinates():
    p1 = CubeVertex(0, (0, 0), (0, 1), (1, 0), 1, 1)
    p2 = CubeVertex(1, (0, 1), (0, 0), (0, 0), 1, 1)
    p3 = CubeVertex(1, (1, 0), (1, 0), (1, 1), 1, 1)
    p4 = CubeVertex(0, (1, 1), (1, 0), (0, 1), 1, 1)

    return p1, p2, p3, p4
    
def plot_rectangle(ax: Axes, p1, p2, p3, p4, color='k', linestyle='-'):
    points = [p.point for p in [p1, p2, p4, p3, p1]]
    x_coords = [pt[0] for pt in points]
    y_coords = [pt[1] for pt in points]
    ax.plot(x_coords, y_coords, color=color, linestyle=linestyle, linewidth=1.2)

def plot(ax: Axes, p1, p2, lim_ranges):
    intersections = get_boundary_intersections(p1, p2)
    
    if len(intersections) == 2:
        ax.plot([intersections[0][0], intersections[1][0]], 
                [intersections[0][1], intersections[1][1]], 'b-')
    
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    perp_slope_p1_p2 = -dx/dy if dy != 0 else float('inf')
    
    for point in [p1, p2]:
        x0, y0 = point
        intersections = []
        
        if np.isinf(perp_slope_p1_p2):  
            intersections.append((x0, -10))
            intersections.append((x0, 10))
        else:
            adj_b = y0 - perp_slope_p1_p2*x0
            
            y = perp_slope_p1_p2*(-10) + adj_b
            if -10 <= y <= 10:
                intersections.append((-10, y))
                
            y = perp_slope_p1_p2*10 + adj_b
            if -10 <= y <= 10:
                intersections.append((10, y))
                
            x = (-10 - adj_b)/perp_slope_p1_p2
            if -10 <= x <= 10:
                intersections.append((x, -10))
                
            x = (10 - adj_b)/perp_slope_p1_p2
            if -10 <= x <= 10:
                intersections.append((x, 10))
        
        if len(intersections) == 2:
            ax.plot([intersections[0][0], intersections[1][0]], 
                    [intersections[0][1], intersections[1][1]], 'r-')
    
    cube_coords = get_cube_coordinates()
    plot_rectangle(ax, *cube_coords)
    

    degrees = plot_vertex_lines(ax, p2, perp_slope_p1_p2, 1)
    
    
    text_str = "\n".join(degrees)
    ax.text(lim_ranges[0], lim_ranges[1], text_str, ha='left', va='top', color="k", size=10)


    ax.plot(p1[0], p1[1], 'go', markersize=5)
    ax.plot(p2[0], p2[1], 'go', markersize=5)
    ax.set_xlim(lim_ranges[0], lim_ranges[1])
    ax.set_ylim(lim_ranges[0], lim_ranges[1])
    
    shift = np.zeros(2) - p1
    
    p2_shift = shift + p2
    
    
    
    angle_point = math.degrees(math.atan2(p2[1], p2[0]))
    angle_shift = math.degrees(math.atan2(p2_shift[1], p2_shift[0]))
    
    ax.set_title(f"angle_point: {round(angle_point, 1)}\nangle_shift: {round(angle_shift, 1)}", fontsize=8)
    print(f"adj_slope: {perp_slope_p1_p2}")

    ax.grid(True)
    ax.set_aspect('equal', adjustable='box')

fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(7, 7), layout="tight")

print(np.shape(axs.flatten()))
ax1, ax2, ax3, ax4 = axs.flatten()
#ax1, ax2, ax3, ax4, ax5, ax6 = axs.flatten()

lim_ranges = -1.5, 3

p1 = (-0.9, 1.1)
p2 = (0.8, -1.)



plot(ax1, p1, p2, lim_ranges)

p1 = (-0.2, 1.3)
p2 = (-0.7, -0.8)



plot(ax2, p1, p2, lim_ranges)

p1 = (-0.2, 1.05)
p2 = (1.8, 0.95)
plot(ax3, p1, p2, lim_ranges)

p1 = (-0.8, 0.6)
p2 = (1.8, 1.3)
plot(ax4, p1, p2, lim_ranges)

"""p1 = (-0.8, 0.6)
p2 = (1.5, 1.5)
plot(ax5, p1, p2, lim_ranges)

p1 = (-0.8, 0.6)
p2 = (0.8, 1.8)
plot(ax6, p1, p2, lim_ranges)
"""

plt.subplots_adjust(wspace=0, hspace=0)


plt.show()


#%%
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import numpy as np

# Cube bounds
x_min, x_max = 0, 2
y_min, y_max = 0, 2
z_min, z_max = 0, 2

# Vertices
vertices = np.array([
    [x_min, y_min, z_min],
    [x_max, y_min, z_min],
    [x_max, y_max, z_min],
    [x_min, y_max, z_min],
    [x_min, y_min, z_max],
    [x_max, y_min, z_max],
    [x_max, y_max, z_max],
    [x_min, y_max, z_max],
])

# Faces
faces = [
    [vertices[i] for i in [0, 1, 2, 3]],
    [vertices[i] for i in [4, 5, 6, 7]],
    [vertices[i] for i in [0, 1, 5, 4]],
    [vertices[i] for i in [2, 3, 7, 6]],
    [vertices[i] for i in [1, 2, 6, 5]],
    [vertices[i] for i in [0, 3, 7, 4]],
]

# Define worldRange from cube bounds
worldRange = np.array([
    [x_min, x_max],
    [y_min, y_max],
    [z_min, z_max],
])

# Compute axis-aligned unit vectors scaled by cube dimension
axis_vectors = [
    np.array([worldRange[0][1] - worldRange[0][0], 0, 0]),
    np.array([0, worldRange[1][1] - worldRange[1][0], 0]),
    np.array([0, 0, worldRange[2][1] - worldRange[2][0]]),
]


scale = 0.25
scaled_vectors = [v * scale for v in axis_vectors]
# Previous code remains the same until the plotting section...
# Previous code remains the same until the plotting section...

fig = plt.figure(figsize=(8, 6))
ax = fig.add_subplot(111, projection='3d')
prism = Poly3DCollection(faces, edgecolors='black', linewidths=1, alpha=0.25)
ax.add_collection3d(prism)

# Define colors for different cross products
cross_colors = ['red', 'purple']  # red for i×j, purple for j×i

for k in range(3):  # Main axis loop
    i, j = [d for d in range(3) if d != k]  # Get perpendicular axes
    
    
   # if i > 0:
    #    continue
   # if j != 2:
   #     continue
    
    
    for vertex in vertices:
        vec_i = scaled_vectors[i]
        vec_j = scaled_vectors[j]
        
        # Compute both cross products
        cross_ij = np.cross(vec_i, vec_j)
        cross_ji = np.cross(vec_j, vec_i)
        
        
        
        
        side_cube = vertex[k]
        
        k_range = worldRange[k]
        
        other_side = k_range[k_range != side_cube].item()
        center_direction = other_side - side_cube
        center_direction = center_direction / abs(center_direction)
        
        cross_ij_dir = cross_ij[k] / abs(cross_ij[k])
        cross_ji_dir = cross_ji[k] / abs(cross_ji[k])
        
        
        
        
        cross_ij = cross_ij / np.linalg.norm(cross_ij) * scale * 1.5
        cross_ji = cross_ji / np.linalg.norm(cross_ji) * scale * 1.5
        
        if cross_ij_dir != center_direction:
            cross_ij[k] *=-1
        if cross_ji_dir != center_direction:
            cross_ji[k] *=-1

            
        x_cross, y_cross, z_cross = zip(vertex, vertex + cross_ij)
        ax.plot(x_cross, y_cross, z_cross, color=cross_colors[0], linewidth=3, alpha=0.7)
            
        label_pos_ij = vertex + cross_ij * 1.1
        ax.text(label_pos_ij[0], label_pos_ij[1], label_pos_ij[2], 
            f"{['X','Y','Z'][i]}×{['X','Y','Z'][j]}", 
            color=cross_colors[0], fontsize=8)

        x_cross, y_cross, z_cross = zip(vertex, vertex + cross_ji)
        ax.plot(x_cross, y_cross, z_cross, color=cross_colors[1], linewidth=3, alpha=0.7)
        
        label_pos_ji = vertex + cross_ji * 1.1
        ax.text(label_pos_ji[0], label_pos_ji[1], label_pos_ji[2], 
            f"{['X','Y','Z'][j]}×{['X','Y','Z'][i]}", 
            color=cross_colors[1], fontsize=8)

# Axes config
ax.set_xlim([x_min, x_max])
ax.set_ylim([y_min, y_max])
ax.set_zlim([z_min, z_max])
ax.set_xlabel('X')
ax.set_ylabel('Y')
ax.set_zlabel('Z')
ax.set_box_aspect([x_max - x_min, y_max - y_min, z_max - z_min])
plt.tight_layout()
plt.title('Cross Products in Both Directions')
plt.show()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option, ABCInstrument, Derivative
    from instruments.utils import InstrumentManager
    from ...data_classes.classes import Slice
    
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import Qt
from . import stylesheets
from datetime import datetime  
import pandas as pd 
from ...settings.widgets import CenteredPopupComboBox
from pprint import pprint
from . import utils as utils_omon



class CustomDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, table):
        super().__init__()
        self.table = table

    def paint(self, painter, option, index):
        HIGHLIGHT_COLOR = QtGui.QColor("#9e9e9f")
        if option.state & QtWidgets.QStyle.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight().color())
        else:
            
            item = self.table.item(index.row(), index.column())
            if item:
                painter.fillRect(option.rect, item.background().color())
        item = self.table.item(index.row(), index.column())
        if item:
            painter.save()
            painter.setFont(item.font())
            text_color = item.foreground().color()
            if option.state & QtWidgets.QStyle.State_Selected:
                text_color = option.palette.highlightedText().color()
            painter.setPen(text_color)
            text_rect = option.rect.adjusted(4, 0, -4, 0)
            painter.drawText(text_rect, item.textAlignment() | Qt.AlignVCenter, item.text())
            painter.restore()

    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        if option.state & QtWidgets.QStyle.State_Selected:
            option.backgroundBrush = option.palette.highlight()
            option.palette.setColor(QtGui.QPalette.HighlightedText, index.data(Qt.ForegroundRole).color())
                
                
class StrikeCellItem(QtWidgets.QTableWidgetItem):   
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(self.flags() | Qt.ItemIsSelectable)
        self.setFlags(self.flags() & ~Qt.ItemIsEditable)
        self.setBackground(QtGui.QBrush("black"))
        self.setForeground(QtGui.QBrush("white"))
        self.setTextAlignment(Qt.AlignCenter)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)


class OptionMetricCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(self.flags() | Qt.ItemIsSelectable)
        self.setFlags(self.flags() & ~Qt.ItemIsEditable)
        self.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)  
        self.setBackground(QtGui.QBrush(QtGui.QColor("black"))) 
        self.setForeground(QtGui.QBrush(QtGui.QColor("white"))) 
        self.setTextAlignment(Qt.AlignRight)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)
        

class BlankCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(self.flags() & ~Qt.ItemIsSelectable)
        self.setFlags(self.flags() & ~Qt.ItemIsEditable)
        self.setBackground(QtGui.QBrush(QtGui.QColor("#414141")))
        self.setForeground(QtGui.QBrush(QtGui.QColor("#fb8b1e")))
        self.setTextAlignment(Qt.AlignRight)

class CellItemContainer(QtCore.QObject):
    def __init__(self, *args):
        super().__init__(*args)
        self.container=[]
    
    def addItem(self, item):
        self.container.append(item)
    
    def checkWidths(self, column, table):
        if not table:
            return True
        
        column_width = table.columnWidth(column)
        
        while True:
            _fits=True
            for item in self.container:
                text = item.text()
                font_metrics = QtGui.QFontMetrics(item.font())
                text_width = font_metrics.horizontalAdvance(text)
                if text_width > column_width:
                    _fits=False
            if _fits:
                break
            else:
                for item in self.container:
                    font = item.font()
                    new_pointSize = font.pointSize()-1
                    font.setPointSize(new_pointSize)
                    item.setFont(font)


class OptionNameCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)  
        self.setBackground(QtGui.QBrush("black"))
        self.setForeground(QtGui.QBrush(QtGui.QColor("#fb8b1e")))
        self.setTextAlignment(Qt.AlignLeft)
        self.setFlags(self.flags() | Qt.ItemIsSelectable)
        self.setFlags(self.flags() & ~Qt.ItemIsEditable)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        
        self.setFont(font)
        
        
class OptionExpiryCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        args = list(args)
        self.expiry = args[0]
        if not self.expiry is None:
            dte = int(round((self.expiry - datetime.now().timestamp()) / 3600 / 24))
            dte_str = f"({dte}d)"
            expiry_str = f"{pd.to_datetime(self.expiry, utc=True, unit="s").strftime('%d-%b-%y')} {dte_str}"
        else:
            expiry_str = args[0]
        args[0]=expiry_str
        
        super().__init__(*args, **kwargs)
        self.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        self.setBackground(QtGui.QBrush(QtGui.QColor("#414141")))
        self.setForeground(QtGui.QBrush(QtGui.QColor("white")))
        self.setTextAlignment(Qt.AlignLeft)
        font = QtGui.QFont("Neue Haas Grotesk", 8)
        self.setFont(font)
    
    def update_from_base(self, appended_str):
        dte = int(round((self.expiry - datetime.now().timestamp()) / 3600 / 24))
        dte_str = f"({dte}d)"
        expiry_str = f"{pd.to_datetime(self.expiry, utc=True, unit="s").strftime('%d-%b-%y')} {dte_str}"
        expiry_str = expiry_str + f" {round(appended_str, 2)}"
        self.setText(expiry_str)

    def does_text_fit(self) -> bool:
        table = self.tableWidget()
        if not table:
            return True
        
        column = self.column()
        column_width = table.columnWidth(column)
        
        font_metrics = QtGui.QFontMetrics(self.font())
        text = self.text()
        text_width = font_metrics.horizontalAdvance(text)
        
        padding = 10 
        return (text_width + padding) <= column_width

class TableColumnItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(self.flags() | Qt.ItemIsEnabled)
        self.setBackground(QtGui.QBrush(QtGui.QColor("#414141")))
        self.setForeground(QtGui.QBrush("white"))
        self.setTextAlignment(Qt.AlignCenter)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        font.setBold(True)
        self.setFont(font) 
        
class StrikeOptionsComboBox(CenteredPopupComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.view().setAutoScroll(False)
    #    self.setStyleSheet(stylesheets.get_settings_stylesheets("StrikeOptionsComboBox"))        
     #   font = QtGui.QFont("Neue Haas Grotesk", 12)
     #   self.setFont(font)

            

class RightClickMenu(QtWidgets.QMenu):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hide()
        
    def eventFilter(self, obj, event):
        if event.type() in (QtCore.QEvent.MouseButtonPress,
                            QtCore.QEvent.KeyPress,
                            QtCore.QEvent.FocusOut
                            ):
            self.close()
        return super().eventFilter(obj, event)
    
    def popup(self, q_point, name_metric_map, parent_table):
        if hasattr(self, "greeks_action"):
            self.removeAction(self.greeks_action)
        self._create_greeks(q_point,
                            name_metric_map,
                            parent_table
                            )
        global_pos = self.parent().mapFromGlobal(q_point)
        self.parent().installEventFilter(self)
        print(self.parent())
        super().popup(global_pos)
        
    def _create_greeks(self, q_point, name_metric_map, parent_table):
        self.greeks_action = self.addAction("Greeks")
        self.greeks_action.triggered.connect(lambda : parent_table.create_subtable(q_point, name_metric_map))
        

class MetricTable(QtWidgets.QTableWidget):
    def __init__(self, name_metric_map, getattr_map, label_price_types, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.getattr_map=getattr_map
        self.label_price_types=label_price_types
        self.nrows = 1 + len(self.getattr_map)
        self.ncols=len(self.label_price_types["label"])
        self.cellitem_container={}
        self._on_view=True
        self._underlying_object = name_metric_map["object"]
        self._initLayout()
        self._initContents()
        self.autoFillBackground()
        self.setShowGrid(False)
        self.pop_table=False
        
    def set_on_view(self, flag):
        self._on_view=flag
        
    def _initLayout(self):
        self.verticalHeader().setVisible(True)
        self.horizontalHeader().setVisible(True)
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)
        self.setStyleSheet(stylesheets.get_settings_stylesheets("SubQTableWidget"))   
        
        self.setRowCount(self.nrows)
        self.setColumnCount(self.ncols)        
        
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        
    def _initContents(self):
        for jdx, px_type in enumerate(self.label_price_types["label"]):
            col_name_item = TableColumnItem(px_type)
            font = QtGui.QFont("Neue Haas Grotesk", 12)
            col_name_item.setFont(font)
            self.setHorizontalHeaderItem(jdx, col_name_item)

        underlying_label = TableColumnItem("underlying_px")
        underlying_label.setTextAlignment(QtCore.Qt.AlignLeft)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        underlying_label.setFont(font)
        self.setVerticalHeaderItem(0, underlying_label)
        if self.columnCount() > 1:
            self.setSpan(0, 0, 1, self.columnCount())

        for idx, name in enumerate(self.getattr_map, start=1):
            row_name_item = TableColumnItem(name)
            row_name_item.setTextAlignment(QtCore.Qt.AlignLeft)
            row_name_item.setFont(font)
            self.setVerticalHeaderItem(idx, row_name_item)
    
                    


class OptionMetricsWithTabs(QtWidgets.QWidget):
    def __init__(self, *, name_metric_map, getattr_map, **kwargs):
        super().__init__(**kwargs)
        self.getattr_map=getattr_map
        self.name_metric_map=name_metric_map
        self.instrument_names=list(name_metric_map.keys())
        pprint(getattr_map)
        self.label_ptype_container={name : inner_dict["selectedPriceTypes"] for name, inner_dict in name_metric_map.items()}
        self.nrows = 2 + len(self.getattr_map)
        self.cellItemContainer={}
        self.name_tab_idx_map={}
        self.buttons=[]
        self._initLayout()
        self._initTabs(name_metric_map)
        self._initTableDimensions(name_metric_map)
        self._initCellItems(name_metric_map)
        self.autoFillBackground()
        btn = self.button_group.button(0)
        btn.click()
        self.fitToTable()
        
    def _initLayout(self):
        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)
        self.button_group = QtWidgets.QButtonGroup()
        self.table_stack = QtWidgets.QStackedLayout()
        
        self.table_stack.currentChanged.connect(self.fitToTable)
        
        self.button_layout = FlowLayout()
        self.button_layout.setContentsMargins(0, 0, 0, 0)
        
        self.main_layout.addLayout(self.button_layout)
        self.main_layout.addLayout(self.table_stack)
        
        self.main_layout.setStretch(0, 0) 
        self.main_layout.setStretch(1, 1) 
    

    def _initTabs(self, metric_object_container):
        for idx, name in enumerate(metric_object_container):
            slot=lambda idx, checked: (self.table_stack.setCurrentIndex(idx),
                                       setattr(self.table_stack.itemAt(idx).widget(), "_on_view", True)) if checked \
                                        else setattr(self.table_stack.itemAt(idx).widget(), "_on_view", False)
            
            btn=QtWidgets.QPushButton(name)
            btn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            btn.setSizePolicy(QtWidgets.QSizePolicy.Expanding, 
                              QtWidgets.QSizePolicy.Fixed)
                        
            if len(metric_object_container) > 1:
                btn.customContextMenuRequested.connect(self._open_button_context)
            
            btn.setCheckable(True)
            btn.setChecked(False)
            self.buttons.append(btn)
            
            self.button_layout.addWidget(btn)
            self.button_group.addButton(btn, id=idx) 
            self.button_group.idToggled.connect(slot)
            self.name_tab_idx_map[name]=idx
        self.button_layout.addStretch()
            
    def _popout_table(self, pos, button):
        ticker = button.text()
        table = self.table_stack.itemAt(self.name_tab_idx_map[ticker]).widget()
        table.set_on_view(True)
        idx = self.name_tab_idx_map[ticker]
        
        table.popped=True
        
        self.button_layout.removeWidget(button)
        button.deleteLater()  
        del self.name_tab_idx_map[ticker]
        self.button_group.removeButton(button)
        
        for name, i in list(self.name_tab_idx_map.items()):
            if idx < i:
                button = self.button_group.buttons()[i-1]
                new_i = self.name_tab_idx_map[name] - 1
                self.button_group.setId(button, new_i)
                self.name_tab_idx_map[name]=new_i
 
        self.table_stack.removeWidget(table)
        table.show()
        
    def _open_button_context(self, pos):
        button = self.sender()
        menu = QtWidgets.QMenu(self)
        print_action = menu.addAction("Open in New Window")
        print_action.triggered.connect(lambda: self._popout_table([pos.x(), pos.y()], button))
        menu.exec(button.mapToGlobal(pos))
    
    def _initTableDimensions(self, metric_object_container):
        for name in self.name_tab_idx_map:
            inner_container=metric_object_container[name]
            table = MetricTable(inner_container, self.getattr_map, self.label_ptype_container[name])
            table.set_on_view(False)
            self.table_stack.addWidget(table)
            
    def closeEvent(self, event):
        self.parent().closingSubtable(self)
        super().closeEvent(event)
            
    def _initCellItems(self, metric_object_container):
        print("\n\ninit")
        for name_instrument, idx in self.name_tab_idx_map.items():
            print(f"\n{name_instrument}")
            inner_container=metric_object_container[name_instrument]
            
            instrument_object=inner_container["object"]

            value = instrument_object.underlying_px
            
            underlyingCellItem = OptionMetricCellItem(f"{value:,.2f}")
            underlyingCellItem.setTextAlignment(QtCore.Qt.AlignLeft)
            table = self.table_stack.itemAt(idx).widget()
            print("dimension")
            print(idx)
            print((table.rowCount(), table.columnCount()))
            table.setItem(0, 0, underlyingCellItem)
            metric_map={}
            
            for jdx, px_type in enumerate(self.name_metric_map[name_instrument]["selectedPriceTypes"]["attribute"]):
                print(px_type)
                metric_type_map = {}
                for idx, (metric_label, metric_attr) in enumerate(self.getattr_map.items(), start=1):
                    print(metric_label)
                    value = getattr(instrument_object, metric_attr)(px_type)
                    if metric_label in ["delta", "gamma"]:
                        val_str=f"{value:,.2g}"
                    else:
                        val_str=f"{value:,.2f}"
                    cell_item = OptionMetricCellItem(val_str)
                    metric_type_map[metric_label]=cell_item
                    table.setItem(idx, jdx, cell_item)
                    print(f"{(idx, jdx)} {cell_item.text()}")
                metric_map[px_type] = metric_type_map
            
            self.cellItemContainer[name_instrument]={"underlying_px" : underlyingCellItem,
                                                     "metrics" : metric_map}

    def updateFromOption(self, instrument_object: Option):
        if self.name_tab_idx_map[instrument_object.ticker] == self.table_stack.currentIndex():
            cell_item = self.table_stack.currentWidget().itemAt(0, 0)
            cell_item.setText(f"{instrument_object.underlying_px:,.2f}")

            inner_cellitem_map = self.cellItemContainer[instrument_object.ticker]
            
            for px_type, metric_map in inner_cellitem_map["metrics"].items():
                for name_metric, cell_item in metric_map.items():
                    value = getattr(instrument_object, self.getattr_map[name_metric])(px_type)
                    
                    if abs(value) < 0.01:
                        val_str=f"{value:,.2g}"
                    else:
                        val_str=f"{value:,.2f}"
                    cell_item.setText(val_str)
                
    def fitToTable(self):
        current_table = self.table_stack.currentWidget()
        table_width = (current_table.verticalHeader().width() + 
                        current_table.horizontalHeader().length() + 
                        current_table.frameWidth() * 2)
        
        table_height = (current_table.horizontalHeader().height() + 
                        current_table.verticalHeader().length() + 
                        current_table.frameWidth() * 2)
        
        button_height = 0
        if self.buttons:
            button_height = self.buttons[0].sizeHint().height() * len(self.buttons)
            button_height += self.button_layout.spacing() * (len(self.buttons) - 1)
        
        margins = self.main_layout.contentsMargins()
        
        total_width = max(table_width, self.button_layout.sizeHint().width()) + margins.left() + margins.right()
        total_height = table_height + button_height + margins.top() + margins.bottom()        
        self.setFixedSize(total_width, total_height)
            
    def resizeEvent(self, event):
        if self.parent() and isinstance(self.window(), QtWidgets.QMainWindow):
            window = self.window()
            frame_size = window.frameSize()
            content_size = window.size()
            
            width_diff = frame_size.width() - content_size.width()
            height_diff = frame_size.height() - content_size.height()
            
            window.resize(self.width() + width_diff,
                          self.height() + height_diff
                          )
        super().resizeEvent(event)
        

class FlowLayout(QtWidgets.QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super().__init__(parent)
        
        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)
            
        self.setSpacing(spacing)
        self.itemList = []
        
    def addItem(self, item):
        self.itemList.append(item)
        
    def count(self):
        return len(self.itemList)
        
    def itemAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList[index]
        return None
        
    def takeAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList.pop(index)
        return None
        
    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Orientation(0))
        
    def hasHeightForWidth(self):
        return True
        
    def heightForWidth(self, width):
        return self.doLayout(QtCore.QRect(0, 0, width, 0), True)
        
    def setGeometry(self, rect):
        super().setGeometry(rect)
        self.doLayout(rect, False)
        
    def sizeHint(self):
        return self.minimumSize()
        
    def minimumSize(self):
        size = QtCore.QSize()
        
        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())
            
        margins = self.contentsMargins()
        size += QtCore.QSize(margins.left() + margins.right(), margins.top() + margins.bottom())
        return size
        
    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0
        
        for item in self.itemList:
            spaceX = self.spacing()
            spaceY = self.spacing()
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0
                
            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))
                
            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())
            
        return y + lineHeight - rect.y()
    
    def addStretch(self, stretch=0):
        spacer = QtWidgets.QSpacerItem(0, 0, 
                                       QtWidgets.QSizePolicy.Policy.Expanding, 
                                       QtWidgets.QSizePolicy.Policy.Minimum)
        self.addItem(spacer)
        
    def addWidget(self, widget):
        self.addChildWidget(widget)  
        self.addItem(QtWidgets.QWidgetItem(widget))


from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING

from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DAxisItem
if TYPE_CHECKING:
    from ...quantities.engines import TickEngine, TickEngineManager, MetricAxisEngine
    from ...gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DAxisItem
    from pyvol_terminal.interfaces.volatility_surface.pyvol_GL3DViewWidget import SurfaceViewWidget
    
import pyqtgraph as pg
from pyqtgraph import opengl 
import numpy as np
from ...gl_3D_graphing.graphics_items import GL3DAxisItem
from PySide6 import QtCore, QtGui
from pyqtgraph import opengl

class CustomAxisItem(pg.AxisItem):
    def __init__(self, axis_direction=None, tick_engine: TickEngine=None, *args, **kwargs):
        self.axis_direction=axis_direction
        self.title=None
        self.tick_engine=tick_engine
        super().__init__(*args, **kwargs)
        self.enableAutoSIPrefix(False)

    def tickStrings(self, values, scale, spacing):
        return self.tick_engine.function(values)  
    
class AxisManager(QtCore.QObject):
    def __init__(self,
                 view_widget,
                 tick_engine_manager: TickEngineManager=None,
                 normaliser=None,
                 n_major_ticks=[6, 6, 6],
                 z_offset=0.3,
                 ):
        super().__init__()
        self.view_widget=view_widget
        self.tick_engine_manager=tick_engine_manager
        self.normaliser=normaliser
        self.str_int_map = {"x" : 0,
                            "y" : 1,
                            "z" : [2, 3],
                            }
        
        self.n_major_ticks=n_major_ticks
        
        self.int_str_map = {v: k for k, val in self.str_int_map.items() for v in (val if isinstance(val, list) else [val])}

        self.n_ticks_dict={0 : n_major_ticks[0],
                           1 : n_major_ticks[1],
                           2 : n_major_ticks[2],
                           3 : n_major_ticks[3]}
        self.axis_2D_items: dict[str, List[CustomAxisItem]] = {0 : [],
                                                               1 : [],
                                                               2 : [],
                                                               3 : []
                                                               }              
        self.axis_3D_items: dict[str, GL3DAxisItem] = []
        
        self.colours = {0 : "white",
                        1 : "white",
                        2 : "white",
                        3 : "white"
                        }
        self.num_dir_map = {0 : 0,
                            1 : 1,
                            2 : 2,
                            3 : 2
                            }
        self.axis_x = None  
        self.axis_y = None
        self.axis_z = None
        self.labels=["Strike", "Expiry", "Implied Volatility", "Implied Volatility"]        
        self.label_map = {0 : "Strike",
                          1 : "Expiry",
                          2 : "Implied Volatility"}
        self.offset_ticks = {0 : np.array([-0.2, 0, 0]),
                             1 : np.array([0, 0.2, 0]),
                             2 : np.array([0.2, 0.2, 0]),
                             3 : np.array([-0.2, 0, 0])
                             }
        self.base_ticknum_positions = {}
        self.tick_size_major=1/15
        
        self.axis_perp = {0 : 1, 
                          1 : 0, 
                          2 : 0,
                          3 : 1}
     
        
        self.anchor_map = {0 : True, 
                           1 : False,
                           2 : False,
                           3 : True
                           }
                
        self.axis_origin_container22 = {0 : (0, 0, 0),
                                      1 : (1, 0, 0),
                                      2 : (1, 1, 0),
                                      3 : (0, 0, 0), 
                                      }
        
        self.axis_origin_container = {0 : ("start", "start"),
                                      1 : ("end", "start"),
                                      2 : ("end", "end"),
                                      3 : ("start", "start"), 
                                      }
        self.z_offset=z_offset
        self.extra_offset = {0 : (0., 0., 0.),
                             1 : (0., 0., 0.),
                             2 : (0., z_offset, 0.),
                             3 : (-z_offset, 0., 0.)
                             }    
        self.padding={0 : 0,
                      1 : 0, 
                      2 : z_offset,
                      3 : z_offset
                      }
        self.axis_map = {
                         0 : [{"perpendicular" : 1,
                               "showTicks" : True,
                               "other_pos" : ("start", "start"),
                               },
                              {"perpendicular" : 2,
                               "showTicks" : False,
                               "other_pos" : ("end", "start"),
                               }
                              ],
                         1 : [{"perpendicular" : 0,
                               "showTicks" : True,
                               "other_pos" : ("end", "start"),
                               },
                              {"perpendicular" : 2,
                               "showTicks" : False,
                               "other_pos" : ("start", "end"),
                               }
                              ],
                         2 : [{"perpendicular" : 0,
                               "showTicks" : True,
                               "other_pos" : ("end", "end"),
                               },
                              {"perpendicular" : 1,
                               "showTicks" : True,
                               "other_pos" : ("start", "start"),
                               }
                              ]
                        }
        
        
        self.tick_direction_container = {0 : (0, -1, 0),
                                         1 : (1, 0, 0),
                                         2 : (1, 0, 0),
                                         3 : (0, -1, 0),
                                         }
        self.offset_labels = {0 : 0.1,
                              1 : 0.1,
                              2 : 0.1,
                              3 : 0.1,
                              }
        self._init_axis()
        self.initialised_default=True

    def _init_axis(self):
        
        anchor_right=False
        
        for ax_dir, ax_opt_list in self.axis_map.items():
            
            for ax_opt in ax_opt_list:
                padding = self.z_offset if ax_dir == 2 or ax_opt["other_pos"]==2 else 0
                if ax_opt["showTicks"]:
                    label = self.label_map[ax_dir]
                else:
                    label=None
                axis_item = GL3DAxisItem.CustomAxisItem3D(tick_engine=self.tick_engine_manager.get_engine(ax_dir),
                                                  direction=ax_dir, 
                                                  direction_perp=ax_opt["perpendicular"],
                                                  other_pos=ax_opt["other_pos"],
                                                  unit_size=1,
                                                  nticks=self.n_major_ticks[ax_dir],
                                                  colour=self.colours[ax_dir],
                                                  padding=padding,
                                                  label=label,
                                                  tick_size_major=self.tick_size_major,
                                                  offset_label=self.offset_labels[ax_dir],
                                                  anchor_right=anchor_right,
                                                  font=QtGui.QFont('Neue Haas Grotesk', 10),
                                                  includeGrid=True
                                                  )
                #self.axis_3D_items[axis] = axis_item
                self.axis_3D_items.append(axis_item)
                break
            break

    def _init_axis2(self):
        for axis_num, label in zip([0, 1, 2, 3], self.labels):
            axis_str = self.int_str_map[axis_num]
            axis_direction = self.num_dir_map[axis_num]
            anchor_right = self.anchor_map[axis_num]
            includeGrids=[True, False] if axis_num < 2 else [True,True]
            
            axis_item = GL3DAxisItem.CustomAxisItem3D(self.tick_engine_manager.get_engine(axis_str),
                                                axis_direction, 
                                                self.axis_perp[axis_num],
                                                other_pos=self.axis_origin_container[axis_num],
                                                unit_size=1,
                                                nticks=self.n_major_ticks[axis_num],
                                                colour=self.colours[axis_num],
                                                padding=self.padding[axis_num],
                                                label=label,

                                                tick_size_major=self.tick_size_major,
                                                
                                                offset_label=self.offset_labels[axis_num],
                                                anchor_right=anchor_right,
                                                font=QtGui.QFont('Neue Haas Grotesk', 10),
                                                includeGrids=includeGrids
                                                )
            self.axis_3D_items[axis_num] = axis_item

    def get_label(self,
                  axis_direction: str
                  ) -> str:
        return self.axis_3D_items[axis_direction].labelItem.getLabel()

    def switch_axis(self, axis_direction, axis_label):
        if isinstance(axis_direction, str):
            axis_direction=self.str_int_map[axis_direction]
        self.axis_3D_items[axis_direction].setLabel(axis_label)  
        for axis_item in self.axis_2D_items[axis_direction]:
            axis_item.setLabel(axis_label)
            
    def add_2D_axis_item(self, axis_item, axis_direction):
        self.axis_2D_items[axis_direction].append(axis_item)
        
    def addWidget(self, widget):
        self.view_widget=widget
        if not self.initialised_default:
            self._init_axis()
    
class GridManager:
    def __init__(self,
                 widget: SurfaceViewWidget,
                 n_major_ticks: int,
                 z_offset: float,
                 ) -> GridManager:
        self.widget=widget
        self.size=[1] * 3
        self.n_major_ticks=n_major_ticks
        self.n_minor_ticks=[2 * t  for t in n_major_ticks]
        self.spacing= [size / minor_tick for size, minor_tick in zip(self.size, self.n_minor_ticks)]
        self._major_minor_scale = 0.5
        
   #     self.grid_xy = self._create_grid(QtGui.QVector3D(*self.size), self.spacing, translation=(0.5, 0.5, 0))
        
        offset_xz = z_offset
        offset_yz = -z_offset 
        
        #self.grid_yz = self._create_grid(QtGui.QVector3D(*self.size), self.spacing, rotation=(90, 0, 1, 0), translation=(0 + offset_yz, 0.5, 0.5))
    #    self.grid_xz = self._create_grid(QtGui.QVector3D(*self.size), self.spacing, rotation=(90, 1, 0, 0), translation=(0.5, 1 + offset_xz, 0.5))
        
        if not widget is None:
            self._addGrids(widget)
    
    def addWidget(self,
                  widget: opengl.GLViewWidget
                  ) -> None:
        self.widget=widget
        self._addGrids(self.widget)
    
    def _addGrids(self,
                  widget: opengl.GLViewWidget
                  ) -> None:
     #   widget.addItem(self.grid_xy)
     #   widget.addItem(self.grid_yz)
      #  widget.addItem(self.grid_xz)

        self.widget=widget
        
    def _create_grid(self, size, spacing, rotation=None, translation=None):        
        grid = GL3DAxisItem.CustomGrid(self.n_major_ticks[0],
                               "",
                               size=size)
  #      grid.setSize(*size)
        grid.setSpacing(*spacing)
        if rotation:
            grid.rotate(*rotation)        
        if translation:
            grid.translate(*translation)
        return grid




#%%%

from dataclasses import dataclass, field
from typing import Dict

@dataclass
class MyDataclass:
    _x: Dict[str, str] = field(init=False, default=dict)
    
    def __post_init__(self):
        self._x = {"1": "1"}
    
    def func(self):
        print(self._x)
    
    
mydc = MyDataclass()
mydc.func()
        